import functools
from typing import Dict

from . import interactive_terminal
from . import azure_cli


def azure_configuration() -> Dict[str, str]:
    # Select Azure subscription
    interactive_terminal.terminal_title("Azure subscription")
    subscription = interactive_terminal.user_input_from_list(
        env_var="WEBVIZ_AZURE_SUBSCRIPTION",
        noun="subscription",
        choices=azure_cli.subscriptions(),
    )

    # Select resource group
    interactive_terminal.terminal_title("Azure resource group")
    resource_group = interactive_terminal.user_input_from_list(
        env_var="WEBVIZ_AZURE_RESOURCE_GROUP",
        noun="resource group",
        choices=azure_cli.resource_groups(subscription),
    )

    # Choose app registration
    interactive_terminal.terminal_title("Azure app registration")
    display_name = interactive_terminal.user_input_optional_reuse(
        env_var="WEBVIZ_AZURE_APP_REGISTRATION_DISPLAY_NAME",
        noun="app registration",
        exists_function=azure_cli.existing_app_registration,
    )

    # Choose Azure storage account
    storage_account_exists = functools.partial(
        azure_cli.storage_account_exists,
        subscription=subscription,
        resource_group=resource_group,
    )

    interactive_terminal.terminal_title("Azure storage account")
    storage_account_name = interactive_terminal.user_input_optional_reuse(
        env_var="WEBVIZ_AZURE_STORAGE_ACCOUNT",
        noun="storage account",
        exists_function=storage_account_exists,
        regex="^[a-z0-9]{3,24}$",
    )

    if not storage_account_exists(storage_account_name):
        azure_cli.create_storage_account(
            subscription, resource_group, storage_account_name
        )
        print("✓ Created storage account.")

    # Choose Azure storage container
    storage_container_exists = functools.partial(
        azure_cli.storage_container_exists,
        account_name=storage_account_name,
        subscription=subscription,
        resource_group=resource_group,
    )

    interactive_terminal.terminal_title("Azure storage container")

    storage_container_name = interactive_terminal.user_input_optional_reuse(
        env_var="WEBVIZ_AZURE_STORAGE_CONTAINER",
        noun="storage container",
        exists_function=storage_container_exists,
    )

    if not storage_container_exists(storage_container_name):
        azure_cli.create_storage_container(
            subscription=subscription,
            resource_group=resource_group,
            storage_account=storage_account_name,
            container=storage_container_name,
        )
        print(f"✓ Created storage container '{storage_container_name}'.")

    return {
        "subscription": subscription,
        "resource_group": resource_group,
        "storage_account_name": storage_account_name,
        "storage_account_key_secret": azure_cli.get_storage_account_access_key(
            subscription=subscription,
            resource_group=resource_group,
            account_name=storage_account_name,
        ),
        "storage_container_name": storage_container_name,
        "display_name": display_name,
    }



from typing import List

from dash import Input

from webviz_config import WebvizPluginABC
from webviz_config.utils import StrEnum
from ._views._plot import (
    PlotView,
    PlotViewElement,
    PlotViewSettingsGroup,
    PlotViewElementSettingsGroup,
)
from ._views._table import TableView, TableViewElement, TableViewSettingsGroup
from ._shared_view_elements import TextViewElement
from ._shared_settings import SharedSettingsGroup


class ExampleWlfPlugin(WebvizPluginABC):
    class Ids(StrEnum):
        PLOT_VIEW = "plot-view"
        TABLE_VIEW = "table-view"
        SHARED_SETTINGS = "shared-settings"

    def __init__(self, title: str):
        super().__init__(stretch=True)

        self.data = [(x, x * x) for x in range(0, 10)]
        self.title = title

        self.settings_group = SharedSettingsGroup()
        self.add_shared_settings_group(
            self.settings_group, ExampleWlfPlugin.Ids.SHARED_SETTINGS
        )

        self.add_view(
            PlotView(
                self.data,
                PlotView.Slots(
                    kindness_selector=Input(
                        self.settings_group.component_unique_id(
                            SharedSettingsGroup.Ids.KINDNESS_SELECTOR
                        ).to_string(),
                        "value",
                    ),
                    power_selector=Input(
                        self.settings_group.component_unique_id(
                            SharedSettingsGroup.Ids.POWER_SELECTOR
                        ).to_string(),
                        "value",
                    ),
                ),
            ),
            ExampleWlfPlugin.Ids.PLOT_VIEW,
        )
        self.add_view(
            TableView(
                self.data,
                TableView.Slots(
                    kindness_selector=Input(
                        self.settings_group.component_unique_id(
                            SharedSettingsGroup.Ids.KINDNESS_SELECTOR
                        ).to_string(),
                        "value",
                    ),
                    power_selector=Input(
                        self.settings_group.component_unique_id(
                            SharedSettingsGroup.Ids.POWER_SELECTOR
                        ).to_string(),
                        "value",
                    ),
                ),
            ),
            ExampleWlfPlugin.Ids.TABLE_VIEW,
        )

    @property
    def tour_steps(self) -> List[dict]:
        return [
            {
                "id": self.view(ExampleWlfPlugin.Ids.PLOT_VIEW)
                .view_element(PlotView.Ids.TEXT)
                .component_unique_id(TextViewElement.Ids.TEXT),
                "content": "Greetings from your example plugin.",
            },
            {
                "id": self.settings_group.component_unique_id("kindness-selector"),
                "content": "You can change here if this shall be friendly or not.",
            },
            {
                "id": self.view(ExampleWlfPlugin.Ids.PLOT_VIEW)
                .view_element(PlotView.Ids.PLOT)
                .component_unique_id(PlotViewElement.Ids.GRAPH),
                "content": "Over here you see a plot that shows x² or x³.",
            },
            {
                "id": self.settings_group.component_unique_id(
                    SharedSettingsGroup.Ids.POWER_SELECTOR
                ),
                "content": "You can change here which exponent you prefer.",
            },
            {
                "id": self.view(ExampleWlfPlugin.Ids.PLOT_VIEW)
                .settings_group(PlotView.Ids.PLOT_SETTINGS)
                .component_unique_id(PlotViewSettingsGroup.Ids.COORDINATES_SELECTOR),
                "content": "...and here you can swap the axes.",
            },
            {
                "id": self.view(ExampleWlfPlugin.Ids.PLOT_VIEW)
                .view_element(PlotView.Ids.PLOT)
                .settings_group(PlotViewElement.Ids.PLOT_SETTINGS)
                .component_unique_id(PlotViewElementSettingsGroup.Ids.COLOR_SELECTOR),
                "content": "You can change here which color you prefer.",
            },
            {
                "id": self.view(ExampleWlfPlugin.Ids.TABLE_VIEW)
                .view_element(TableView.Ids.TABLE)
                .component_unique_id(TableViewElement.Ids.TABLE),
                "content": "There is also a table visualizing the data.",
            },
            {
                "id": self.view(ExampleWlfPlugin.Ids.TABLE_VIEW)
                .settings_group(TableView.Ids.TABLE_SETTINGS)
                .component_unique_id(TableViewSettingsGroup.Ids.ORDER_SELECTOR),
                "content": "You can change the order of the table here.",
            },
        ]



from typing import Optional

from dash import html, Dash

from .. import WebvizPluginABC, EncodedFile


class ExampleDataDownload(WebvizPluginABC):
    def __init__(self, app: Dash, title: str):
        super().__init__()

        self.title = title
        self.set_callbacks(app)

    @property
    def layout(self) -> html.H1:
        return html.H1(self.title)

    def set_callbacks(self, app: Dash) -> None:
        @app.callback(self.plugin_data_output, self.plugin_data_requested)
        def _user_download_data(data_requested: bool) -> Optional[EncodedFile]:
            return (
                WebvizPluginABC.plugin_compressed_data(
                    filename="webviz-data.zip",
                    content=[
                        {"filename": "some_file.txt", "content": "Some download data"}
                    ],
                )
                if data_requested
                else None
            )



from typing import Callable, List, Optional, Type, Union
import abc

from dash.development.base_component import Component
import webviz_core_components as wcc

from ._layout_base_abc import LayoutBaseABC
from ._layout_unique_id import LayoutUniqueId


class SettingsGroupABC(LayoutBaseABC):
    def __init__(self, title: str) -> None:
        super().__init__()

        self.title = title
        self._layout_created: bool = False
        self._visible_in_views: List[str] = []
        self._not_visible_in_views: List[str] = []

    def _set_visible_in_views(self, visible_in_views: List[str]) -> None:
        self._visible_in_views = visible_in_views

    def _set_not_visible_in_views(self, not_visible_in_views: List[str]) -> None:
        self._not_visible_in_views = not_visible_in_views

    def _set_plugin_register_id_func(
        self, func: Callable[[Union[str, List[str]]], None]
    ) -> None:
        self._plugin_register_id_func = func

    def register_component_unique_id(self, component_name: str) -> str:
        uuid = self.component_unique_id(component_name).to_string()
        if self._plugin_register_id_func and not self._layout_created:
            self._plugin_register_id_func(uuid)

        return uuid

    def component_unique_id(self, component_name: str) -> LayoutUniqueId:
        component_uuid = LayoutUniqueId(other=self.get_unique_id())
        component_uuid.set_component_id(component_name)
        return component_uuid

    @abc.abstractmethod
    def layout(self) -> Union[List[Component], Type[Component]]:
        raise NotImplementedError

    def _wrapped_layout(
        self,
        view_id: Optional[str] = "",
        plugin_id: Optional[str] = "",
        always_open: bool = False,
    ) -> Type[Component]:
        layout = self.layout()
        wrapped_layout = wcc.WebvizSettingsGroup(
            id=str(self.get_unique_id()),
            title=self.title,
            viewId=view_id,
            pluginId=plugin_id,
            visibleInViews=self._visible_in_views
            if len(self._visible_in_views) > 0
            else None,
            notVisibleInViews=self._not_visible_in_views
            if len(self._not_visible_in_views) > 0
            else None,
            alwaysOpen=always_open,
            children=layout if isinstance(layout, list) else [layout],
        )
        self._layout_created = True
        return wrapped_layout

    def set_callbacks(self) -> None:
        pass



import time
import functools
import secrets
import asyncio
import pathlib
import datetime
import warnings
import webbrowser
from typing import List, Dict, Optional, Tuple

import requests
from tqdm.asyncio import tqdm

try:
    from azure.identity import InteractiveBrowserCredential
    from azure.core.exceptions import HttpResponseError
    from azure.mgmt.resource.subscriptions import SubscriptionClient
    from azure.mgmt.resource import ResourceManagementClient
    from azure.mgmt.storage import StorageManagementClient
    from azure.storage.blob import BlobServiceClient
    from azure.storage.blob.aio import ContainerClient

    AZURE_CLI_INSTALLED = True
except ModuleNotFoundError:
    AZURE_CLI_INSTALLED = False

GRAPH_BASE_URL = "https://graph.microsoft.com"
PIMCOMMON_URL = (
    "https://portal.azure.com/#blade/Microsoft_Azure_PIMCommon/ActivationMenuBlade"
)


@functools.lru_cache
def _credential():  # type: ignore[no-untyped-def]
    if not AZURE_CLI_INSTALLED:
        raise RuntimeError(
            "In order to use webviz deploy features, you need to first install "
            "the optional deploy dependencies. You can do this by e.g. running "
            "'pip install webviz-config[deployment]'"
        )

    return InteractiveBrowserCredential()


def _subscription_id(subscription_name: str = None) -> str:
    subscription_list = SubscriptionClient(_credential()).subscriptions.list()

    if subscription_name is None:
        return next(subscription_list).subscription_id

    for sub in subscription_list:
        if sub.display_name == subscription_name:
            return sub.subscription_id

    raise ValueError(f"Could not find a subscription with name {subscription_name}")


def _connection_string(
    subscription: str, resource_group: str, storage_account: str
) -> str:
    key = get_storage_account_access_key(subscription, resource_group, storage_account)
    return (
        f"DefaultEndpointsProtocol=https;AccountName={storage_account};AccountKey={key}"
    )


def subscriptions() -> List[str]:
    """Returns list of all Azure subscriptions logged in user has read access to."""
    return [
        sub.display_name
        for sub in SubscriptionClient(_credential()).subscriptions.list()
    ]


def resource_groups(subscription: str) -> List[str]:
    """Returns list of all Azure resource group names logged in user has read access to
    within given subscription."""

    rmc = ResourceManagementClient(_credential(), _subscription_id(subscription))
    return [rg.name for rg in rmc.resource_groups.list()]


def storage_account_name_available(name: str) -> Tuple[bool, str]:
    storage_client = StorageManagementClient(_credential(), _subscription_id())
    result = storage_client.storage_accounts.check_name_availability({"name": name})
    return (result.name_available, result.message)


def storage_account_exists(name: str, subscription: str, resource_group: str) -> bool:
    storage_client = StorageManagementClient(
        _credential(), _subscription_id(subscription)
    )

    if any(
        account.name == name
        for account in storage_client.storage_accounts.list_by_resource_group(
            resource_group
        )
    ):
        return True

    for account in storage_client.storage_accounts.list():
        if account.name == name:
            if account.resourceGroup != resource_group:
                warnings.warn(
                    f"Storage account with name {name} found, but it belongs "
                    f"to another resource group ({account.resourceGroup}."
                )
            return True

    return False


def storage_container_exists(
    container_name: str, account_name: str, subscription: str, resource_group: str
) -> bool:
    storage_client = StorageManagementClient(
        _credential(), _subscription_id(subscription)
    )
    containers = storage_client.blob_containers.list(resource_group, account_name)
    return any(container.name == container_name for container in containers)


def create_storage_account(subscription: str, resource_group: str, name: str) -> None:
    """Creates an Azure storage account. Also adds upload access, as well
    as possibility to list/generate access keys, to the user creating it
    (i.e. the currently logged in user).

    Note that Azure documentation states that it can take up to five minutes
    after the command has finished until the added access is enabled in practice.
    """

    storage_client = StorageManagementClient(
        _credential(), _subscription_id(subscription)
    )

    azure_pim_already_open = False

    while True:
        try:
            return storage_client.storage_accounts.begin_create(
                resource_group,
                name,
                {
                    "sku": {"name": "Standard_ZRS"},
                    "kind": "StorageV2",
                    "location": "northeurope",
                    "encryption": {
                        "key_source": "Microsoft.Storage",
                        "services": {"blob": {"key_type": "Account", "enabled": True}},
                    },
                },
            ).result()
        except HttpResponseError as exc:
            if "AuthorizationFailed" in str(exc):
                if not azure_pim_already_open:
                    webbrowser.open(f"{PIMCOMMON_URL}/azurerbac")
                    print(
                        "Not able to create new storage account. Do you have "
                        "enough priviliges to do it? We automatically opened the URL "
                        "to where you activate Azure PIM. Please activate/add necessary "
                        "priviliges."
                    )
                    azure_pim_already_open = True
                print("New attempt of creating storage account in 30 seconds.")
                time.sleep(30)
            else:
                raise RuntimeError("Not able to create new storage account.") from exc


def get_storage_account_access_key(
    subscription: str,
    resource_group: str,
    account_name: str,
) -> str:
    storage_client = StorageManagementClient(
        _credential(), _subscription_id(subscription)
    )

    azure_pim_already_open = False

    while True:
        try:
            return (
                storage_client.storage_accounts.list_keys(resource_group, account_name)
                .keys[0]
                .value
            )
        except HttpResponseError as exc:
            if "AuthorizationFailed" in str(exc):
                if not azure_pim_already_open:
                    webbrowser.open(f"{PIMCOMMON_URL}/azurerbac")
                    print(
                        "Not able to get access keys for storage account. Do you have "
                        "enough priviliges to do it? We automatically opened the URL "
                        "to where you activate Azure PIM. Please activate/add necessary "
                        "priviliges. If not inherited from higher privileges, you will "
                        "as a minmium need role 'Storage Account Key Operator Service Role' "
                        f"for the storage account {account_name}."
                    )
                    azure_pim_already_open = True
                print("New attempt of getting storage account keys in 30 seconds.")
                time.sleep(30)
            else:
                raise RuntimeError("Not able to get storage account keys.") from exc


def create_storage_container(
    subscription: str,
    resource_group: str,
    storage_account: str,
    container: str,
) -> None:
    BlobServiceClient.from_connection_string(  # type: ignore[attr-defined]
        _connection_string(subscription, resource_group, storage_account)
    ).get_container_client(container).create_container()


def storage_container_upload_folder(
    subscription: str,
    resource_group: str,
    storage_name: str,
    container_name: str,
    source_folder: pathlib.Path,
) -> None:

    paths_to_upload = [path for path in source_folder.rglob("*") if path.is_file()]

    async def _upload_file(
        container_client: ContainerClient,
        path: pathlib.Path,
        source_folder: pathlib.Path,
        semaphore: asyncio.Semaphore,
    ) -> None:
        async with semaphore:
            with open(path, "rb") as file_handle:
                await container_client.upload_blob(
                    name=path.relative_to(source_folder).as_posix(),
                    data=file_handle,
                    overwrite=True,
                )

    async def _upload_blob() -> None:
        async with ContainerClient.from_connection_string(  # type: ignore[attr-defined]
            _connection_string(subscription, resource_group, storage_name),
            container_name,
        ) as container_client:

            class UploadTasks:
                def __init__(
                    self, paths_to_upload: List[pathlib.Path], max_open_files: int = 100
                ):
                    self.index = 0
                    self.paths_to_upload = paths_to_upload
                    self.semaphore = asyncio.Semaphore(max_open_files)

                # Forward reference not possible on Python 3.6:
                def __iter__(self):  # type: ignore[no-untyped-def]
                    return self

                def __next__(self) -> asyncio.Task:
                    try:
                        task = asyncio.create_task(
                            _upload_file(
                                container_client,
                                self.paths_to_upload[self.index],
                                source_folder,
                                self.semaphore,
                            )
                        )
                    except IndexError as exc:
                        raise StopIteration from exc
                    self.index += 1
                    return task

                def __len__(self) -> int:
                    return len(self.paths_to_upload)

            for task in tqdm.as_completed(
                UploadTasks(paths_to_upload),
                bar_format="{l_bar} {bar} | Uploaded {n_fmt}/{total_fmt}",
            ):
                await task

    asyncio.run(_upload_blob())


def _graph_headers() -> Dict[str, str]:
    token = _credential().get_token(f"{GRAPH_BASE_URL}/.default").token
    return {"Authorization": f"bearer {token}", "Content-type": "application/json"}


def _object_id_from_app_id(app_registration_id: str) -> str:
    endpoint = f"applications?$filter=appID eq '{app_registration_id}'"

    data = requests.get(
        f"{GRAPH_BASE_URL}/v1.0/{endpoint}",
        headers=_graph_headers(),
    ).json()

    object_id = data["value"][0]["id"]
    return object_id


def existing_app_registration(display_name: str) -> Optional[str]:
    """Returns application (client) ID with given display_name if it exists,
    otherwise returns None.
    """
    endpoint = f"applications?$filter=displayName eq '{display_name}'"
    data = requests.get(
        f"{GRAPH_BASE_URL}/v1.0/{endpoint}", headers=_graph_headers()
    ).json()

    if data["value"]:
        return data["value"][0]["appId"]
    return None


def create_app_registration(display_name: str) -> str:

    existing_app_id = existing_app_registration(display_name)
    if existing_app_id is not None:
        return existing_app_id

    data = requests.post(
        f"{GRAPH_BASE_URL}/v1.0/applications",
        json={"displayName": display_name, "signInAudience": "AzureADMyOrg"},
        headers=_graph_headers(),
    ).json()

    if "error" in data and data["error"]["code"] == "Authorization_RequestDenied":
        raise PermissionError("Insufficient privileges to create new app registration.")

    app_id = data["appId"]

    webbrowser.open(
        "https://portal.azure.com/#blade/Microsoft_AAD_RegisteredApps/"
        f"ApplicationMenuBlade/Manifest/appId/{app_id}"
    )

    input(
        """
    Due to current limitations in Microsoft Graph API
    (https://docs.microsoft.com/en-us/azure/active-directory/develop/reference-app-manifest),
    we need to update the app manifest manually. In the browser tab that was opened,
    please update "oauth2Permissions" to this:

    	"oauth2Permissions": [
		{
			"adminConsentDescription": "Access on behalf of the signed-in user",
			"adminConsentDisplayName": "Access on behalf of the signed-in user",
			"id": "00000000-0000-0000-0000-000000000001",
			"isEnabled": true,
			"lang": null,
			"origin": "Application",
			"type": "User",
			"userConsentDescription": "Access application on your behalf",
			"userConsentDisplayName": "Access application on your behalf",
			"value": "user_impersonation"
		}
	],

    Press enter to continue."""
    )

    return app_id


def create_secret(app_registration_id: str, secret_description: str) -> str:

    object_id = _object_id_from_app_id(app_registration_id)

    end_datetime = datetime.datetime.now() + datetime.timedelta(days=365)
    data = requests.post(
        f"{GRAPH_BASE_URL}/v1.0/applications/{object_id}/addPassword",
        json={
            "passwordCredential": {
                "displayName": secret_description,
                "endDateTime": end_datetime.isoformat(),
            }
        },
        headers=_graph_headers(),
    ).json()

    return data["secretText"]


def add_reply_url(app_registration_id: str, reply_url: str) -> None:
    """Will add web reply url to given app registration id, if it does not alredy exist."""

    object_id = _object_id_from_app_id(app_registration_id)

    data = requests.get(
        f"{GRAPH_BASE_URL}/v1.0/applications/{object_id}",
        headers=_graph_headers(),
    ).json()

    web = data["web"]

    if reply_url not in web["redirectUris"]:
        web["redirectUris"].append(reply_url)

    requests.patch(
        f"{GRAPH_BASE_URL}/v1.0/applications/{object_id}",
        json={"web": web},
        headers=_graph_headers(),
    )


def create_service_principal(app_registration_id: str) -> Tuple[str, str]:

    endpoint = f"servicePrincipals?$filter=appID eq '{app_registration_id}'"
    data = requests.get(
        f"{GRAPH_BASE_URL}/v1.0/{endpoint}",
        headers=_graph_headers(),
    ).json()

    if "error" not in data and data["value"]:
        data_object = data["value"][0]
        if not data_object["appRoleAssignmentRequired"]:
            raise RuntimeError(
                "Service principal already exists, and it does not require app role "
                "assignments. Deployment stopped, as this should be set to true in "
                "order to secure access."
            )
    else:
        data_object = requests.post(
            f"{GRAPH_BASE_URL}/v1.0/servicePrincipals",
            json={"appId": app_registration_id, "appRoleAssignmentRequired": True},
            headers=_graph_headers(),
        ).json()

        if "error" in data:
            raise RuntimeError(f"Graph query failed with response {data}")

    object_id = data_object["id"]
    directory_tenant_id = data_object["appOwnerOrganizationId"]

    return object_id, directory_tenant_id


def azure_app_registration_setup(
    display_name: str, proxy_redirect_url: str
) -> Dict[str, str]:

    azure_pim_already_open = False

    while True:
        try:
            app_registration_id = create_app_registration(display_name)
            object_id, tenant_id = create_service_principal(app_registration_id)
            break
        except PermissionError:
            if not azure_pim_already_open:
                webbrowser.open(f"{PIMCOMMON_URL}/aadmigratedroles")
                azure_pim_already_open = True

                print(
                    "Not able to create new app registration. Do you have enough "
                    "priviliges to do it? We automatically opened the URL where you "
                    "can activate Azure PIM. Please activate necessary priviliges."
                )

            print("New attempt of app registration in 30 seconds.")
            time.sleep(30)

    proxy_client_secret = create_secret(app_registration_id, "cli secret")
    add_reply_url(app_registration_id, proxy_redirect_url)

    return {
        "app_registration_id": app_registration_id,
        "object_id": object_id,
        "proxy_client_secret": proxy_client_secret,
        "proxy_cookie_secret": secrets.token_urlsafe(nbytes=16),
        "proxy_redirect_url": proxy_redirect_url,
        "tenant_id": tenant_id,
    }



from typing import Optional

import pytest

from webviz_config.webviz_factory import WebvizFactory
from webviz_config.webviz_factory_registry import WebvizFactoryRegistry


# pylint: disable=no-self-use
class FactoryA(WebvizFactory):
    def create_obj_a(self) -> str:
        return "A"


class FactoryB(WebvizFactory):
    def create_obj_b(self) -> str:
        return "B"


class FactoryBSub(FactoryB):
    def create_obj_b_sub(self) -> str:
        return "B_sub"


# pylint: disable=too-few-public-methods
class UnrelatedFactory:
    pass


def create_initialized_registry() -> WebvizFactoryRegistry:
    registry = WebvizFactoryRegistry()
    registry.initialize({"MyFactory": "someValue"})
    return registry


def test_uninitialized_access() -> None:
    registry = WebvizFactoryRegistry()

    with pytest.raises(RuntimeError):
        _settings = registry.all_factory_settings


def test_initialization_and_basic_access() -> None:
    registry = create_initialized_registry()

    settings = registry.all_factory_settings
    assert "MyFactory" in settings

    factory = registry.get_factory(FactoryA)
    assert factory is None


def test_multiple_initializations() -> None:
    registry = WebvizFactoryRegistry()
    registry.initialize({"MyFactory": "someValue"})

    with pytest.raises(RuntimeError):
        registry.initialize({"MyFactory": "someValue"})


def test_set_get_factory() -> None:
    registry = create_initialized_registry()

    factory_a = FactoryA()
    registry.set_factory(FactoryA, factory_a)

    factory_b = FactoryB()
    registry.set_factory(FactoryB, factory_b)

    f_a: Optional[FactoryA] = registry.get_factory(FactoryA)
    assert f_a is factory_a

    f_b: Optional[FactoryB] = registry.get_factory(FactoryB)
    assert f_b is factory_b


def test_set_get_derived_factory() -> None:
    registry = create_initialized_registry()

    factory_b = FactoryB()
    factory_b_sub = FactoryBSub()

    registry.set_factory(FactoryB, factory_b)
    registry.set_factory(FactoryBSub, factory_b_sub)
    assert registry.get_factory(FactoryB) is factory_b
    assert registry.get_factory(FactoryBSub) is factory_b_sub

    # This should be fine
    registry.set_factory(FactoryB, factory_b_sub)
    assert registry.get_factory(FactoryB) is factory_b_sub

    # This should not be legal, but cannot currently figure out how
    # to disallow it using type hinting. Currently throws an exception
    with pytest.raises(TypeError):
        registry.set_factory(FactoryBSub, factory_b)


def test_set_mismatched_factory_types() -> None:
    registry = create_initialized_registry()

    factory_a: FactoryA = FactoryA()
    factory_b: FactoryB = FactoryB()
    factory_b_sub: FactoryBSub = FactoryBSub()

    # These four are legal
    registry.set_factory(FactoryA, factory_a)
    registry.set_factory(FactoryB, factory_b)
    registry.set_factory(FactoryB, factory_b_sub)
    registry.set_factory(FactoryBSub, factory_b_sub)

    # These are rightly caught as errors by type checking since they don't derive from WebvizFactory
    # registry.set_factory(UnrelatedFactory, UnrelatedFactory())
    # registry.set_factory(FactoryA, UnrelatedFactory())
    # registry.set_factory(FactoryA, 123)

    # The following setter calls are illegal, but can't figure out how
    # to get type hinting to disallow them
    # For now, we rely on exceptions
    with pytest.raises(TypeError):
        registry.set_factory(FactoryA, factory_b)
    with pytest.raises(TypeError):
        registry.set_factory(FactoryBSub, factory_b)


def test_get_mismatched_factory_types() -> None:
    registry = create_initialized_registry()

    _f_a: Optional[FactoryA] = None
    _f_b: Optional[FactoryB] = None
    _f_b_sub: Optional[FactoryBSub] = None

    # These four are legal and do indeed pass wrt type checking
    _f_a = registry.get_factory(FactoryA)
    _f_b = registry.get_factory(FactoryB)
    _f_b = registry.get_factory(FactoryBSub)
    _f_b_sub = registry.get_factory(FactoryBSub)

    # These are rightly caught as errors by type checking, good
    # _f_b = registry.get_factory(FactoryA)
    # _f_b_sub = registry.get_factory(FactoryB)



from typing import Callable, cast, Dict, List, Optional, Type, Union
from enum import Enum
from uuid import uuid4

from dash import Input, Output
from dash.development.base_component import Component
import webviz_core_components as wcc

from ._settings_group_abc import SettingsGroupABC
from ._layout_base_abc import LayoutBaseABC
from ._layout_unique_id import LayoutUniqueId


class NoParentPlugin(Exception):
    pass


class UnknownId(Exception):
    pass


class ViewElementABC(LayoutBaseABC):
    def __init__(
        self, flex_grow: int = 1, screenshot_filename: str = "webviz-screenshot.png"
    ) -> None:
        super().__init__()

        self._flex_grow = flex_grow
        self._screenshot_filename = screenshot_filename
        self._add_download_button = False

        self._settings: List[SettingsGroupABC] = []
        self._layout_created: bool = False

    def _set_plugin_register_id_func(
        self, func: Callable[[Union[str, List[str]]], None]
    ) -> None:
        super()._set_plugin_register_id_func(func)

        for setting in self._settings:
            # pylint: disable=protected-access
            setting._set_plugin_register_id_func(func)

    def _set_plugin_get_store_unique_id_func(self, func: Callable[[str], str]) -> None:
        super()._set_plugin_get_store_unique_id_func(func)
        # pylint: disable=protected-access
        for setting in self._settings:
            setting._set_plugin_get_store_unique_id_func(func)

    def _set_unique_id(self, parent_unique_id: LayoutUniqueId) -> None:
        super()._set_unique_id(parent_unique_id)

        for setting in self._settings:
            # pylint: disable=protected-access
            setting._set_unique_id(self.get_unique_id())

    def set_flex_grow(self, flex_grow: int) -> None:
        self._flex_grow = flex_grow

    def register_component_unique_id(self, component_name: str) -> str:
        uuid = self.component_unique_id(component_name).to_string()
        if self._plugin_register_id_func and not self._layout_created:
            self._plugin_register_id_func(uuid)

        return uuid

    def component_unique_id(self, component_name: str) -> LayoutUniqueId:
        component_unique_id = LayoutUniqueId(other=self.get_unique_id())
        component_unique_id.set_component_id(component_name)
        return component_unique_id

    def add_settings_group(
        self, settings_group: SettingsGroupABC, settings_group_id: str
    ) -> None:
        # pylint: disable=protected-access
        settings_group.get_unique_id().set_settings_group_id(settings_group_id)
        settings_group._set_unique_id(self._unique_id)
        self._settings.append(settings_group)

    def setting_group_unique_id(
        self, settings_id: str, element: Optional[str] = None
    ) -> str:
        setting = next(
            (
                el
                for el in self.settings_groups()
                if el.get_unique_id().get_settings_group_id() == settings_id
            ),
            None,
        )
        if not setting:
            available_ids = [
                cast(str, el.get_unique_id().get_settings_group_id())
                for el in self.settings_groups()
                if el.get_unique_id().get_settings_group_id() is not None
            ]
            raise UnknownId(
                f"Could not find settings group with id '{settings_id}'.\n"
                f"Available ids are: {' ,'.join(available_ids)}"
            )

        uuid = LayoutUniqueId(other=setting.get_unique_id())
        if element:
            uuid.set_component_id(element)
        return uuid.to_string()

    def settings_group(self, settings_group_id: str) -> SettingsGroupABC:
        settings_group = next(
            (
                el
                for el in self.settings_groups()
                if el.get_unique_id().get_settings_group_id() == settings_group_id
            ),
            None,
        )
        if settings_group:
            return settings_group

        raise LookupError(
            f"""Invalid settings group id: '{settings_group_id}.
            Available settings group ids: {[
                el.get_unique_id().get_settings_group_id()
                for el in self.settings_groups()
            ]}
            """
        )

    def view_element_data_output(self) -> Output:
        self._add_download_button = True
        return Output(str(self.get_unique_id().to_string()), "download")

    def view_element_data_requested(self) -> Input:
        return Input(str(self.get_unique_id().to_string()), "data_requested")

    def _wrapped_layout(self) -> Union[str, Type[Component]]:
        layout = self.inner_layout()
        self._layout_created = True
        return layout

    def outer_layout(self) -> Type[Component]:
        # pylint: disable=protected-access
        return wcc.WebvizViewElement(
            id=str(self.get_unique_id()),
            showDownload=self._add_download_button,
            flexGrow=self._flex_grow,
            children=[
                self._wrapped_layout(),
                *[
                    setting._wrapped_layout(always_open=True)
                    for setting in self.settings_groups()
                ],
            ],
        )

    def inner_layout(self) -> Union[str, Type[Component]]:
        raise NotImplementedError

    def settings_groups(self) -> List[SettingsGroupABC]:
        return self._settings

    def _set_all_callbacks(self) -> None:
        for setting in self._settings:
            setting.set_callbacks()

        self.set_callbacks()

    def set_callbacks(self) -> None:
        pass


class LayoutElementType(Enum):
    ROW = 1
    COLUMN = 2


class ViewLayoutElement(LayoutBaseABC):
    def __init__(
        self,
        layout_element_type: LayoutElementType,
        parent_view: "ViewABC",
        flex_grow: int = 1,
    ):
        super().__init__()

        self._parent_view: ViewABC = parent_view
        self._children: List[Union[ViewLayoutElement, ViewElementABC]] = []
        self._flex_grow = flex_grow
        self.__type = layout_element_type

    def make_row(
        self, row_id: Optional[str] = None, flex_grow: int = 1
    ) -> "ViewLayoutElement":
        # pylint: disable=protected-access
        if row_id is None:
            row_id = str(uuid4())
        row = ViewLayoutElement(LayoutElementType.ROW, self._parent_view, flex_grow)
        row._set_unique_id(self._unique_id)
        row.get_unique_id().set_view_element_id(row_id)
        self._children.append(row)
        self._parent_view._add_layout_element(row, row_id)
        return row

    def make_column(
        self, column_id: Optional[str] = None, flex_grow: int = 1
    ) -> "ViewLayoutElement":
        # pylint: disable=protected-access
        if column_id is None:
            column_id = str(uuid4())
        column = ViewLayoutElement(
            LayoutElementType.COLUMN, self._parent_view, flex_grow
        )
        column._set_unique_id(self._unique_id)
        column.get_unique_id().set_view_element_id(column_id)
        self._children.append(column)
        self._parent_view._add_layout_element(column, column_id)
        return column

    def add_row(self, row: "ViewLayoutElement", row_id: Optional[str] = None) -> None:
        # pylint: disable=protected-access
        row_id = row_id if row_id is not None else str(uuid4())
        row._set_unique_id(self.get_unique_id())
        row.get_unique_id().set_view_element_id(row_id)
        self._parent_view._add_layout_element(row, row_id)
        self._children.append(row)

    def add_column(
        self, column: "ViewLayoutElement", column_id: Optional[str] = None
    ) -> None:
        # pylint: disable=protected-access
        column_id = column_id if column_id is not None else str(uuid4())
        column._set_unique_id(self.get_unique_id())
        column.get_unique_id().set_view_element_id(column_id)
        self._parent_view._add_layout_element(column, column_id)
        self._children.append(column)

    def add_view_element(
        self, view_element: ViewElementABC, view_element_id: str, flex_grow: int = 1
    ) -> None:
        # pylint: disable=protected-access
        view_element.set_flex_grow(flex_grow)
        self._parent_view._add_view_element(view_element, view_element_id)
        self._children.append(view_element)

    def add_view_elements(self, view_elements: Dict[str, ViewElementABC]) -> None:
        for view_element_id, view_element in view_elements.items():
            self.add_view_element(view_element, view_element_id)

    def _set_all_callbacks(self) -> None:
        for child in self._children:
            # pylint: disable=protected-access
            child._set_all_callbacks()

    def _set_plugin_register_id_func(
        self, func: Callable[[Union[str, List[str]]], None]
    ) -> None:
        super()._set_plugin_register_id_func(func)
        for element in self._children:
            # pylint: disable=protected-access
            element._set_plugin_register_id_func(func)

    def _set_plugin_get_store_unique_id_func(self, func: Callable[[str], str]) -> None:
        super()._set_plugin_get_store_unique_id_func(func)
        # pylint: disable=protected-access
        for element in self._children:
            element._set_plugin_get_store_unique_id_func(func)

    def _set_unique_id(self, parent_unique_id: LayoutUniqueId) -> None:
        # pylint: disable=protected-access
        super()._set_unique_id(parent_unique_id)

        for child in self._children:
            child._set_unique_id(parent_unique_id)

    @property
    def layout(self) -> Type[Component]:
        # pylint: disable=protected-access
        if self.__type == LayoutElementType.ROW:
            return wcc.WebvizPluginLayoutRow(
                id=str(self.get_unique_id()),
                flexGrow=self._flex_grow,
                children=[
                    wcc.WebvizViewElement(
                        id=str(el.get_unique_id()),
                        # pylint: disable=protected-access
                        showDownload=el._add_download_button,
                        flexGrow=el._flex_grow,
                        children=[
                            el._wrapped_layout(),
                            *[
                                setting._wrapped_layout(always_open=True)
                                for setting in el.settings_groups()
                            ],
                        ],
                    )
                    if isinstance(el, ViewElementABC)
                    else el.layout
                    for el in self._children
                ],
            )
        return wcc.WebvizPluginLayoutColumn(
            id=str(self.get_unique_id()),
            flexGrow=self._flex_grow,
            children=[
                wcc.WebvizViewElement(
                    id=str(el.get_unique_id()),
                    showDownload=el._add_download_button,
                    flexGrow=el._flex_grow,
                    children=[
                        el._wrapped_layout(),
                        *[
                            setting._wrapped_layout(always_open=True)
                            for setting in el.settings_groups()
                        ],
                    ],
                )
                if isinstance(el, ViewElementABC)
                else el.layout
                for el in self._children
            ],
        )


# pylint: disable=too-many-public-methods
class ViewABC(LayoutBaseABC):
    def __init__(self, name: str) -> None:
        super().__init__()
        self.name = name

        self._add_download_button = False
        self._child_elements: List[Union[ViewElementABC, ViewLayoutElement]] = []
        self._view_elements: List[ViewElementABC] = []
        self._layout_elements: List[ViewLayoutElement] = []
        self._settings_groups: List[SettingsGroupABC] = []
        self._plugin_register_id_func: Optional[
            Callable[[Union[str, List[str]]], None]
        ] = None
        self._get_plugin_shared_settings: Optional[
            Callable[[], List[SettingsGroupABC]]
        ] = None

    def shared_settings_group(self, settings_group_id: str) -> SettingsGroupABC:
        if not self._get_plugin_shared_settings:
            raise NoParentPlugin(
                f"The view {self.name} has not been added to a plugin yet."
            )
        settings_group = next(
            (
                el
                for el in self._get_plugin_shared_settings()
                if el.get_unique_id().get_settings_group_id() == settings_group_id
            ),
            None,
        )
        if settings_group:
            return settings_group

        raise LookupError(
            f"""Invalid shared settings group id: '{settings_group_id}.
            Available shared settings group ids: {[
                el.get_unique_id().get_settings_group_id()
                for el in self._get_plugin_shared_settings()
            ]}
            """
        )

    def shared_settings_group_unique_id(
        self, settings_id: str, element: Optional[str] = None
    ) -> str:
        if not self._get_plugin_shared_settings:
            raise NoParentPlugin(
                f"The view {self.name} has not been added to a plugin yet."
            )
        setting = next(
            (
                el
                for el in self._get_plugin_shared_settings()
                if el.get_unique_id().get_settings_group_id() == settings_id
            ),
            None,
        )
        if not setting:
            available_ids = [
                cast(str, el.get_unique_id().get_settings_group_id())
                for el in self._get_plugin_shared_settings()
                if el.get_unique_id().get_settings_group_id() is not None
            ]
            raise UnknownId(
                f"Could not find settings group with id '{settings_id}'.\n"
                f"Available ids are: {' ,'.join(available_ids)}"
            )

        uuid = LayoutUniqueId(other=setting.get_unique_id())
        if element:
            uuid.set_component_id(element)
        return uuid.to_string()

    def _set_get_plugin_shared_settings_func(
        self, func: Callable[[], List[SettingsGroupABC]]
    ) -> None:
        self._get_plugin_shared_settings = func

    def _set_plugin_register_id_func(
        self, func: Callable[[Union[str, List[str]]], None]
    ) -> None:
        # pylint: disable=protected-access
        self._plugin_register_id_func = func
        for element in self._child_elements:
            element._set_plugin_register_id_func(func)

        for setting in self._settings_groups:
            setting._set_plugin_register_id_func(func)

    def _set_plugin_get_store_unique_id_func(self, func: Callable[[str], str]) -> None:
        super()._set_plugin_get_store_unique_id_func(func)
        # pylint: disable=protected-access
        for element in self._child_elements:
            element._set_plugin_get_store_unique_id_func(func)

        for setting in self._settings_groups:
            setting._set_plugin_get_store_unique_id_func(func)

    def _set_unique_id(self, parent_unique_id: LayoutUniqueId) -> None:
        # pylint: disable=protected-access
        super()._set_unique_id(parent_unique_id)

        for element in self._child_elements:
            element._set_unique_id(self.get_unique_id())

        for setting in self._settings_groups:
            setting._set_unique_id(self.get_unique_id())

    def unique_id(self, element: Optional[str] = None) -> str:
        if element:
            return f"{element}-{self._unique_id}"
        return str(self._unique_id)

    def view_element_unique_id(
        self, view_element_id: str, element: Optional[str] = None
    ) -> str:
        view_element = next(
            (
                el
                for el in self.view_elements()
                if el.get_unique_id().get_view_element_id() == view_element_id
            ),
            None,
        )
        if not view_element:
            available_ids = [
                cast(str, el.get_unique_id().get_view_element_id())
                for el in self.view_elements()
                if el.get_unique_id().get_view_element_id() is not None
            ]
            raise UnknownId(
                f"Could not find view element with id '{view_element_id}'.\n"
                f"Available ids are: {' ,'.join(available_ids)}"
            )

        uuid = LayoutUniqueId(other=view_element.get_unique_id())
        if element:
            uuid.set_component_id(element)
        return uuid.to_string()

    def view_element(self, view_element_id: str) -> "ViewElementABC":
        view_element = next(
            (
                el
                for el in self.view_elements()
                if el.get_unique_id().get_view_element_id() == view_element_id
            ),
            None,
        )
        if view_element:
            return view_element

        raise LookupError(
            f"""Invalid view element id: '{view_element_id}.
            Available view element ids: {[
                el.get_unique_id().get_view_element_id()
                for el in self.view_elements()
            ]}
            """
        )

    def layout_element(self, layout_element_id: str) -> "ViewLayoutElement":
        layout_element = next(
            (
                el
                for el in self.layout_elements()
                if el.get_unique_id().get_view_element_id() == layout_element_id
            ),
            None,
        )
        if layout_element:
            return layout_element

        raise LookupError(
            f"""Invalid layout element id: '{layout_element_id}.
            Available layout element ids: {[
                el.get_unique_id().get_view_element_id()
                for el in self.layout_elements()
            ]}
            """
        )

    def settings_group(self, settings_group_id: str) -> SettingsGroupABC:
        settings_group = next(
            (
                el
                for el in self.settings_groups()
                if el.get_unique_id().get_settings_group_id() == settings_group_id
            ),
            None,
        )
        if settings_group:
            return settings_group

        raise LookupError(
            f"""Invalid settings group id: '{settings_group_id}.
            Available settings group ids: {[
                el.get_unique_id().get_settings_group_id()
                for el in self.settings_groups()
            ]}
            """
        )

    def settings_group_unique_id(
        self, settings_id: str, element: Optional[str] = None
    ) -> str:
        setting = next(
            (
                el
                for el in self.settings_groups()
                if el.get_unique_id().get_settings_group_id() == settings_id
            ),
            None,
        )
        if not setting:
            available_ids = [
                cast(str, el.get_unique_id().get_settings_group_id())
                for el in self.settings_groups()
                if el.get_unique_id().get_settings_group_id() is not None
            ]
            raise UnknownId(
                f"Could not find settings group with id '{settings_id}'.\n"
                f"Available ids are: {' ,'.join(available_ids)}"
            )

        uuid = LayoutUniqueId(other=setting.get_unique_id())
        if element:
            uuid.set_component_id(element)
        return uuid.to_string()

    def add_view_element(
        self, view_element: ViewElementABC, view_element_id: str
    ) -> None:
        # pylint: disable=protected-access
        view_element._set_unique_id(self._unique_id)
        view_element.get_unique_id().set_view_element_id(view_element_id)
        self._child_elements.append(view_element)
        self._view_elements.append(view_element)

    def _add_view_element(
        self, view_element: ViewElementABC, view_element_id: str
    ) -> None:
        # pylint: disable=protected-access
        view_element._set_unique_id(self._unique_id)
        view_element.get_unique_id().set_view_element_id(view_element_id)
        self._view_elements.append(view_element)

    def _add_layout_element(
        self, layout_element: ViewLayoutElement, layout_element_id: str
    ) -> None:
        # pylint: disable=protected-access
        layout_element._set_unique_id(self._unique_id)
        layout_element.get_unique_id().set_view_element_id(layout_element_id)
        self._layout_elements.append(layout_element)

    def add_row(
        self, layout_element_id: Optional[str] = None, flex_grow: int = 1
    ) -> ViewLayoutElement:
        row = ViewLayoutElement(LayoutElementType.ROW, self, flex_grow)
        layout_element_id = (
            layout_element_id if layout_element_id is not None else str(uuid4())
        )
        row.get_unique_id().set_view_element_id(layout_element_id)
        self._layout_elements.append(row)
        self._child_elements.append(row)
        return row

    def add_column(
        self, layout_element_id: Optional[str] = None, flex_grow: int = 1
    ) -> ViewLayoutElement:
        column = ViewLayoutElement(LayoutElementType.COLUMN, self, flex_grow)
        layout_element_id = (
            layout_element_id if layout_element_id is not None else str(uuid4())
        )
        column.get_unique_id().set_view_element_id(layout_element_id)
        self._layout_elements.append(column)
        self._child_elements.append(column)
        return column

    def add_view_elements(self, view_elements: Dict[str, ViewElementABC]) -> None:
        for view_element_id, view_element in view_elements.items():
            self.add_view_element(view_element, view_element_id)

    def remove_view_element(self, view_element_id: str) -> bool:
        old_length = len(self._view_elements)
        self._view_elements = [
            view_element
            for view_element in self._view_elements
            if view_element.get_unique_id().get_view_element_id() == view_element_id
        ]
        return old_length != len(self._view_elements)

    def add_settings_group(
        self, settings_group: SettingsGroupABC, settings_group_id: str
    ) -> None:
        # pylint: disable=protected-access
        settings_group._set_unique_id(self._unique_id)
        settings_group.get_unique_id().set_settings_group_id(settings_group_id)
        self._settings_groups.append(settings_group)

    def add_settings_groups(self, settings_groups: Dict[str, SettingsGroupABC]) -> None:
        for settings_group_id, settings_group in settings_groups.items():
            self.add_settings_group(settings_group, settings_group_id)

    def _set_all_callbacks(self) -> None:
        # pylint: disable=protected-access
        for element in self._view_elements:
            element._set_all_callbacks()

        for setting in self._settings_groups:
            setting.set_callbacks()

        self.set_callbacks()

    def view_elements(self) -> List[ViewElementABC]:
        return self._view_elements

    def layout_elements(self) -> List[ViewLayoutElement]:
        return self._layout_elements

    def settings_groups(self) -> List[SettingsGroupABC]:
        return self._settings_groups

    def view_data_output(self) -> Output:
        self._add_download_button = True
        return Output(str(self.get_unique_id().to_string()), "download")

    def view_data_requested(self) -> Input:
        return Input(str(self.get_unique_id().to_string()), "data_requested")

    def inner_layout(self) -> List[Type[Component]]:
        return [
            el.outer_layout() if isinstance(el, ViewElementABC) else el.layout
            for el in self._child_elements
        ]

    def outer_layout(self) -> Type[Component]:
        return wcc.WebvizView(
            id=str(self.get_unique_id()),
            children=self.inner_layout(),
        )

    def set_callbacks(self) -> None:
        pass



from webviz_config._docs._build_docs import _split_docstring


def test_split_docstring():
    tests = [
        (
            "This is a test with only description.",
            ["This is a test with only description."],
        ),
        (
            (
                " This is a test with only description, "
                "but which has leading and trailing spaces.  "
            ),
            [
                (
                    "This is a test with only description, "
                    "but which has leading and trailing spaces."
                )
            ],
        ),
        (
            "Test with some newlines\n\n  \n    and a 4 space indent",
            ["Test with some newlines\n\n\nand a 4 space indent"],
        ),
        (
            "Test with a \n    4 space and a \n  2 space indent\n",
            ["Test with a \n  4 space and a \n2 space indent"],
        ),
        (
            "Test with a \n    4 space and a \n  2 space indent\n",
            ["Test with a \n  4 space and a \n2 space indent"],
        ),
        (
            (
                " This is a test with description, arguments and data input."
                "\n\n    ---\n\n    The docstring is indented by 4 spaces:\n    "
                "* The argument list has a sub list indented by 8 spaces\n        like this."
                "\n    ---\n    The data input section is not very interesting."
            ),
            [
                "This is a test with description, arguments and data input.\n",
                (
                    "\nThe docstring is indented by 4 spaces:\n"
                    "* The argument list has a sub list indented by 8 spaces\n    like this."
                ),
                "The data input section is not very interesting.",
            ],
        ),
        (
            (
                "\tThis is a test with description, arguments and data input,"
                " where indents are made with tabs and not spaces-"
                "\n\n\t---\n\n\tThe docstring is indented by 1 tab:\n"
                "\t* The argument list has a sub list indented by 2 tabs\n\t\tlike this."
                "\n\t---\n\tThe data input section is not very interesting."
            ),
            [
                (
                    "This is a test with description, arguments and data input,"
                    " where indents are made with tabs and not spaces-\n"
                ),
                (
                    "\nThe docstring is indented by 1 tab:\n"
                    "* The argument list has a sub list indented by 2 tabs\n\tlike this."
                ),
                "The data input section is not very interesting.",
            ],
        ),
        (
            "This is a test with only description, which includes a linebreak.\n",
            ["This is a test with only description, which includes a linebreak."],
        ),
    ]
    for test in tests:
        assert _split_docstring(test[0]) == test[1]



import json
import shutil
import tempfile
import subprocess
from pathlib import Path


def binary_available() -> bool:
    """Returns True if the GitHub CLI binary (gh) is
    availabile in $PATH, otherwise returns False.
    """
    return shutil.which("gh") is not None


def logged_in() -> bool:
    # pylint: disable=subprocess-run-check
    result = subprocess.run(
        ["gh", "auth", "status"],
        capture_output=True,
    )
    return b"Logged in to github.com" in result.stderr


def log_in() -> None:
    # pylint: disable=subprocess-run-check
    subprocess.run(
        ["gh", "auth", "login"],
    )


def repo_exists(github_slug: str) -> bool:
    # pylint: disable=subprocess-run-check
    result = subprocess.run(
        ["gh", "repo", "view", github_slug],
        capture_output=True,
    )

    if not result.stderr:
        return True
    if b"Could not resolve" in result.stderr:
        return False
    raise RuntimeError(result.stderr)


def create_github_repository(github_slug: str, directory: Path) -> Path:
    """Creates a new private github repository. github_slug is on format
    owner/reponame.
    """
    if not "/" in github_slug:
        raise ValueError("repo_path argument should be on format owner/reponame")

    subprocess.run(
        ["gh", "repo", "create", github_slug, "--private", "--confirm"],
        capture_output=True,
        check=True,
        cwd=directory,
    )

    return directory / github_slug.split("/")[1]


def turn_on_github_vulnerability_alers(directory: Path) -> None:
    subprocess.run(
        [
            "gh",
            "api",
            "repos/:owner/:repo/vulnerability-alerts",
            "--method",
            "PUT",
            "--header",
            "Accept: application/vnd.github.dorian-preview+json",
        ],
        check=True,
        cwd=directory,
    )


def _call_post_api(endpoint: str, data: dict, directory: Path) -> None:
    subprocess.run(
        [
            "gh",
            "api",
            endpoint,
            "--method",
            "POST",
            "--input",
            "-",
            "--silent",
        ],
        input=json.dumps(data),
        check=True,
        cwd=directory,
        text=True,
    )


def add_webhook(directory: Path, receiver_url: str, secret: str) -> None:
    data = {
        "name": "web",
        "active": True,
        "config": {
            "url": receiver_url,
            "content_type": "json",
            "secret": secret,
        },
    }
    _call_post_api(data=data, endpoint="repos/:owner/:repo/hooks", directory=directory)


def add_deploy_key(directory: Path, title: str, key: str) -> None:
    _call_post_api(
        data={"title": title, "key": key},
        endpoint="repos/:owner/:repo/keys",
        directory=directory,
    )


def read_file_in_repository(github_slug: str, filename: str) -> str:
    with tempfile.TemporaryDirectory() as tmp_dir:
        temp_dir = Path(tmp_dir)
        subprocess.run(
            ["git", "clone", f"git@github.com:{github_slug}"],
            check=True,
            cwd=temp_dir,
            capture_output=True,
        )
        clone_path = temp_dir / github_slug.split("/")[1]

        return (clone_path / Path(filename)).read_text()


def commit_portable_webviz(
    github_slug: str,
    source_directory: Path,
    commit_message: str = "Initial commit",
    branch_name: str = "main",
) -> None:

    with tempfile.TemporaryDirectory() as tmp_dir:
        temp_dir = Path(tmp_dir)
        subprocess.run(
            ["git", "clone", f"git@github.com:{github_slug}"],
            check=True,
            cwd=temp_dir,
            capture_output=True,
        )

        clone_path = temp_dir / github_slug.split("/")[1]

        shutil.copytree(
            source_directory,
            clone_path,
            dirs_exist_ok=True,
            ignore=shutil.ignore_patterns("resources"),
        )

        commands = [
            ["git", "add", "."],
            ["git", "commit", "-m", commit_message, "--allow-empty"],
            ["git", "branch", "-M", branch_name],
            ["git", "push", "-u", "origin", branch_name],
        ]
        for command in commands:
            subprocess.run(
                command,
                check=True,
                cwd=clone_path,
                capture_output=True,
            )



from typing import cast

import pytest

from webviz_config import WebvizConfigTheme, WebvizSettings


def test_construction_and_basic_access() -> None:
    # pylint: disable=unidiomatic-typecheck
    the_shared_settings = {"somenumber": 10, "somestring": "abc"}
    the_theme = WebvizConfigTheme("dummyThemeName")
    settings_obj = WebvizSettings(the_shared_settings, the_theme)

    copy_of_shared_settings = settings_obj.shared_settings
    assert copy_of_shared_settings is not the_shared_settings
    assert type(copy_of_shared_settings) == type(the_shared_settings)
    assert copy_of_shared_settings == the_shared_settings
    the_shared_settings["somestring"] = "MODIFIED"
    assert copy_of_shared_settings != the_shared_settings

    copy_of_theme = settings_obj.theme
    assert copy_of_theme is not the_theme
    assert type(copy_of_theme) == type(the_theme)
    assert copy_of_theme.__dict__ == the_theme.__dict__
    the_theme.theme_name = "MODIFIED"
    assert copy_of_theme.__dict__ != the_theme.__dict__


def test_construction_with_invalid_types() -> None:
    with pytest.raises(TypeError):
        theme = WebvizConfigTheme("dummyThemeName")
        _settings_obj = WebvizSettings(cast(dict, None), theme)

    with pytest.raises(TypeError):
        shared_settings = {"somenumber": 10, "somestring": "abc"}
        _settings_obj = WebvizSettings(shared_settings, cast(WebvizConfigTheme, None))



from pathlib import Path
from typing import List

from dash import dcc

from .. import WebvizPluginABC
from ..webviz_store import webvizstore


class SyntaxHighlighter(WebvizPluginABC):
    """Adds support for syntax highlighting of code. Language is automatically detected.

---

* **`filename`:** Path to a file containing the code to highlight.
* **`dark_theme`:** If `True`, the code is shown with a dark theme. Default is \
                `False`, giving a light theme.
"""

    def __init__(self, filename: Path, dark_theme: bool = False):

        super().__init__()

        self.filename = filename
        self.config = {"theme": "dark"} if dark_theme else {"theme": "light"}

    def add_webvizstore(self) -> List[tuple]:
        return [(get_path, [{"path": self.filename}])]

    @property
    def layout(self) -> dcc.Markdown:
        return dcc.Markdown(
            f"```\n{get_path(self.filename).read_text()}\n```",
            highlight_config=self.config,
        )


@webvizstore
def get_path(path: Path) -> Path:
    return path



import json
import webbrowser
from typing import Optional

from ._user_data_dir import user_data_dir
from .themes import installed_themes

USER_SETTINGS_FILE = user_data_dir() / "user_settings.json"


def set_user_preferences(
    theme: Optional[str] = None, browser: Optional[str] = None
) -> None:

    preferences = (
        json.loads(USER_SETTINGS_FILE.read_text())
        if USER_SETTINGS_FILE.is_file()
        else {}
    )

    new_preferences = {}

    if theme is not None:
        if theme not in installed_themes:
            raise ValueError(
                f"Theme {theme} is not one of the installed themes ({', '.join(installed_themes)})"
            )
        new_preferences["theme"] = theme

    if browser is not None:
        try:
            webbrowser.get(using=browser)
        except webbrowser.Error as exc:
            raise ValueError(
                f"Could not find an installed browser with the name {browser}."
            ) from exc

        new_preferences["browser"] = browser

    if new_preferences:
        preferences.update(new_preferences)
        USER_SETTINGS_FILE.parent.mkdir(parents=True, exist_ok=True)
        USER_SETTINGS_FILE.write_text(json.dumps(preferences))


def get_user_preference(setting: str) -> Optional[str]:
    return (
        json.loads(USER_SETTINGS_FILE.read_text()).get(setting)
        if USER_SETTINGS_FILE.is_file()
        else None
    )



import types
import typing
import inspect

from .._plugin_abc import WebvizPluginABC


def _get_webviz_plugins(module: types.ModuleType) -> list:
    """Returns a list of all Webviz plugins
    in the module given as input.
    """

    def _is_webviz_plugin(obj: typing.Any) -> bool:
        return (
            inspect.isclass(obj)
            and issubclass(obj, WebvizPluginABC)
            and obj is not WebvizPluginABC
        )

    return inspect.getmembers(module, _is_webviz_plugin)



import webviz_config
from webviz_config.plugins import PLUGIN_PROJECT_METADATA


def test_webviz_config_metadata():

    metadata = PLUGIN_PROJECT_METADATA["webviz-config"]

    assert metadata["dist_version"] == webviz_config.__version__
    assert metadata["documentation_url"] == "https://equinor.github.io/webviz-config"
    assert metadata["download_url"] == "https://pypi.org/project/webviz-config"
    assert metadata["source_url"] == "https://github.com/equinor/webviz-config"
    assert metadata["tracker_url"] == "https://github.com/equinor/webviz-config/issues"

    assert "dash" in metadata["dependencies"]  # dash is a direct dependency
    assert "flask" in metadata["dependencies"]  # flask is an indirect dependency



"""These are the basic Webviz configuration plugins, distributed through
the utility itself.
"""

from importlib.metadata import distributions

from ._utils import load_webviz_plugins_with_metadata, PluginProjectMetaData


PLUGIN_METADATA, PLUGIN_PROJECT_METADATA = load_webviz_plugins_with_metadata(
    distributions(), globals()
)



import sys
import getpass
import datetime
import pathlib
import argparse
from typing import Dict, Tuple

import yaml
import jinja2

from ._config_parser import ConfigParser


def write_script(
    args: argparse.Namespace,
    build_directory: pathlib.Path,
    template_filename: str,
    output_filename: str,
) -> Tuple[set, Dict[str, dict]]:
    """Writes rendered script to build directory. Also returns information regarding
    assets and which plugins that are incluced in the user provided configuration file.
    """

    config_parser = ConfigParser(args.yaml_file)
    configuration = config_parser.configuration

    configuration.update(
        {
            "author": getpass.getuser(),
            "config_folder": repr(args.yaml_file.resolve().parent),
            "current_date": datetime.date.today().strftime("%Y-%m-%d"),
            "debug": args.debug,
            "loglevel": args.loglevel,
            "portable": args.portable is not None,
            "shared_settings": config_parser.shared_settings,
            "sys_executable": sys.executable,
            "theme_name": args.theme,
        }
    )

    if args.logconfig is not None:
        configuration["logging_config_dict"] = yaml.safe_load(
            args.logconfig.read_text()
        )

    template_environment = jinja2.Environment(  # nosec
        loader=jinja2.PackageLoader("webviz_config", "templates"),
        undefined=jinja2.StrictUndefined,
        autoescape=False,
    )

    template = template_environment.get_template(template_filename)

    (build_directory / output_filename).write_text(template.render(configuration))

    return config_parser.assets, config_parser.plugin_metadata



import re
import sys
import pathlib
import inspect
from typing import Dict, List, Optional, Any
import warnings

import yaml

import webviz_config.plugins
from .utils import terminal_colors
from .utils._get_webviz_plugins import _get_webviz_plugins
from . import _deprecation_store as _ds

SPECIAL_ARGS = ["self", "app", "webviz_settings", "_call_signature"]


def _call_signature(
    plugin_name: str,
    kwargs: dict,
    config_folder: pathlib.Path,
    contact_person: Optional[dict] = None,
) -> tuple:
    # pylint: disable=too-many-branches,too-many-statements,too-many-locals
    """Takes as input the name of a plugin together with user given arguments
    (originating from the configuration file). Returns the equivalent Python code wrt.
    initiating an instance of that plugin (with the given arguments).

    Raises ParserError in the following scenarios:
      * User is missing a required (i.e. no default value) __init__ argument
      * User provided an argument not existing in the class' __init__ function
      * User has given one of the protected arguments in SPECIAL_ARGS
      * If there is type mismatch between user given argument value, and type
        hint in __init__ signature (given that type hint exist)
    """
    argspec = inspect.getfullargspec(
        getattr(webviz_config.plugins, plugin_name).__init__
    )

    if argspec.defaults is not None:
        required_args = argspec.args[: -len(argspec.defaults)]
    else:
        required_args = argspec.args

    for arg in required_args:
        if arg not in SPECIAL_ARGS and arg not in kwargs:
            raise ParserError(
                f"{terminal_colors.RED}{terminal_colors.BOLD}"
                f"The plugin `{plugin_name}` requires the argument "
                f"`{arg}` in your configuration file. "
                "Run the command `webviz docs` if you want "
                "to see documentation of the plugin."
                f"{terminal_colors.END}"
            )

    for arg in list(kwargs):
        if arg in SPECIAL_ARGS:
            raise ParserError(
                f"{terminal_colors.RED}{terminal_colors.BOLD}"
                f"Argument `{arg}` not allowed."
                f"{terminal_colors.END}"
            )

        if arg == "contact_person":
            if not isinstance(kwargs["contact_person"], dict):
                raise ParserError(
                    f"{terminal_colors.RED}{terminal_colors.BOLD}"
                    f"The contact information provided for "
                    f"`{plugin_name}` is not a dictionary. "
                    f"{terminal_colors.END}"
                )

            if any(
                key not in ["name", "phone", "email"]
                for key in kwargs["contact_person"]
            ):
                raise ParserError(
                    f"{terminal_colors.RED}{terminal_colors.BOLD}"
                    f"Unrecognized contact information key given to `{plugin_name}`. "
                    f"Should be 'name', 'phone' and/or 'email'."
                    f"{terminal_colors.END}"
                )

            contact_person = kwargs.pop("contact_person")

        elif arg not in argspec.args:
            raise ParserError(
                f"{terminal_colors.RED}{terminal_colors.BOLD}"
                f"Unrecognized argument. `{plugin_name}` "
                f"does not take an argument `{arg}`. "
                "Run the command `webviz docs` if you want "
                "to see documentation of the plugin."
                f"{terminal_colors.END}"
            )

        if arg in argspec.annotations:
            expected_type = argspec.annotations[arg]

            if expected_type == pathlib.Path:
                kwargs[arg] = (config_folder / pathlib.Path(kwargs[arg])).resolve()
            elif expected_type == List[pathlib.Path]:
                kwargs[arg] = [
                    (config_folder / pathlib.Path(patharg)).resolve()
                    for patharg in kwargs[arg]
                ]
            try:
                if not isinstance(kwargs[arg], expected_type):
                    raise ParserError(
                        f"{terminal_colors.RED}{terminal_colors.BOLD}"
                        f"The value provided for argument `{arg}` "
                        f"given to `{plugin_name}` is "
                        f"of type `{type(kwargs[arg]).__name__}`. "
                        f"Expected type "
                        f"`{argspec.annotations[arg].__name__}`. "
                        "Run the command `webviz docs` if you want "
                        "to see documentation of the plugin."
                        f"{terminal_colors.END}"
                    )
            # We currently do not support subscriptable types, e.g. Optional[], List[], Dict[] etc.
            # This will be supported after https://github.com/equinor/webviz-config/issues/199
            except TypeError:
                pass

    kwargs_including_defaults = kwargs
    plugin_deprecation_warnings = []
    argument_deprecation_warnings = []

    deprecated_plugin = _ds.DEPRECATION_STORE.get_stored_plugin_deprecation(
        getattr(webviz_config.plugins, plugin_name)
    )
    if deprecated_plugin:
        plugin_deprecation_warnings.append(deprecated_plugin.short_message)
        warnings.warn(
            f"""Plugin '{plugin_name}' has been deprecated.
------------------------
{deprecated_plugin.short_message}
===
{deprecated_plugin.long_message}
""",
            FutureWarning,
        )

    deprecations = _ds.DEPRECATION_STORE.get_stored_plugin_argument_deprecations(
        getattr(webviz_config.plugins, plugin_name).__init__
    )

    signature = inspect.signature(getattr(webviz_config.plugins, plugin_name).__init__)
    for key, value in signature.parameters.items():
        if value.default is not inspect.Parameter.empty and key not in kwargs.keys():
            kwargs_including_defaults[key] = value.default

    for deprecation in deprecations:
        if isinstance(deprecation, _ds.DeprecatedArgument):
            if deprecation.argument_name in kwargs_including_defaults.keys():
                argument_deprecation_warnings.append(deprecation.short_message)
                warnings.warn(
                    f"Deprecated Argument: '{deprecation.argument_name}' with value "
                    f"'{kwargs_including_defaults[deprecation.argument_name]}' "
                    f"in plugin '{plugin_name}\n'"
                    "------------------------\n"
                    f"{deprecation.short_message}\n"
                    "===\n"
                    f"{deprecation.long_message}",
                    FutureWarning,
                )
        elif isinstance(deprecation, _ds.DeprecatedArgumentCheck):
            mapped_args: Dict[str, Any] = {}
            for arg in deprecation.argument_names:
                for name, value in kwargs_including_defaults.items():
                    if arg == name:
                        mapped_args[arg] = value
                        break

            result = deprecation.callback(**mapped_args)
            if result:
                argument_deprecation_warnings.append(result[0])
                _values = [
                    value
                    for key, value in kwargs_including_defaults.items()
                    if key in deprecation.argument_names
                ]
                warnings.warn(
                    f"Deprecated Argument(s): '{deprecation.argument_names}' "
                    f"with value(s) '{_values}' in plugin '{plugin_name}'\n"
                    "------------------------\n"
                    f"{result[0]}\n"
                    "===\n"
                    f"{result[1]}",
                    FutureWarning,
                )

    special_args = ""
    if "app" in argspec.args:
        special_args += "app=app, "
    if "webviz_settings" in argspec.args:
        special_args += "webviz_settings=webviz_settings, "

    return (
        f"{plugin_name}({special_args}**{kwargs})",
        (
            f"plugin_layout(contact_person={contact_person}"
            f", plugin_deprecation_warnings={plugin_deprecation_warnings}"
            f", argument_deprecation_warnings={argument_deprecation_warnings})"
        ),
    )


class ParserError(Exception):
    pass


class ConfigParser:

    STANDARD_PLUGINS = [
        name for (name, _) in _get_webviz_plugins(webviz_config.plugins)
    ]

    def __init__(self, yaml_file: pathlib.Path):

        ConfigParser.check_for_tabs_in_file(yaml_file)

        try:
            self._configuration = yaml.safe_load(yaml_file.read_text())
        except yaml.MarkedYAMLError as excep:
            extra_info = (
                f"There is something wrong in the configuration file {yaml_file}. "
            )

            # Looking forward to only supporting Python 3.8+ and PEP572...
            problem_mark = getattr(excep, "problem_mark", None)
            if problem_mark is not None:
                extra_info += (
                    "The typo is probably somewhere around "
                    f"line {problem_mark.line + 1}."
                )

            raise type(excep)(
                f"{excep}. {terminal_colors.RED}{terminal_colors.BOLD}"
                f"{extra_info}{terminal_colors.END}"
            ).with_traceback(sys.exc_info()[2])

        self._config_folder = pathlib.Path(yaml_file).parent
        self._page_titles: List[str] = []
        self._page_ids: List[str] = []
        self._assets: set = set()
        self._plugin_metadata: Dict[str, dict] = {}
        self._used_plugin_packages: set = set()
        self.clean_configuration()

    @staticmethod
    def check_for_tabs_in_file(path: pathlib.Path) -> None:

        line_numbers_with_tabs = sorted(
            list(
                {
                    i + 1
                    for i, line in enumerate(path.read_text().splitlines())
                    if "\t" in line
                }
            )
        )

        if line_numbers_with_tabs:
            raise ParserError(
                f"{terminal_colors.RED}{terminal_colors.BOLD}"
                "The configuration file contains tabs. You should use ordinary spaces "
                "instead. The following lines contain tabs:\n"
                f"{', '.join([str(line_no) for line_no in line_numbers_with_tabs])}"
                f"{terminal_colors.END}"
            )

    def _generate_page_id(self, title: str) -> str:
        """From the user given title, this function provides a unique
        human readable page id, not already present in `self._page_ids`.
        """

        base_id = re.sub("[^-a-z0-9_]+", "", title.lower().replace(" ", "-"))

        page_id = base_id

        count = 1
        while page_id in self._page_ids:
            count += 1
            page_id = f"{base_id}{count}"

        return page_id

    def clean_configuration(self) -> None:
        """Various cleaning and checks of the raw configuration read
        from the user provided yaml configuration file.
        """

        self.configuration["_imports"] = set()

        if "shared_settings" in self.configuration:
            self._shared_settings = self.configuration["shared_settings"]
        else:
            self._shared_settings = {}

        if "title" not in self.configuration:
            self.configuration["title"] = "Webviz - Powered by Dash"

        if "pages" not in self.configuration and "layout" not in self.configuration:
            raise ParserError(
                f"{terminal_colors.RED}{terminal_colors.BOLD}"
                "The configuration file does not have "
                "information regarding which pages to create."
                f"{terminal_colors.END}"
            )

        if "pages" in self.configuration and not isinstance(
            self.configuration["pages"], list
        ):
            raise ParserError(
                f"{terminal_colors.RED}{terminal_colors.BOLD}"
                "The configuration input belonging to the "
                "`pages` keyword should be a list."
                f"{terminal_colors.END}"
            )

        if "pages" in self.configuration:
            warnings.warn(
                "The configuration keyword `pages` is going to be deprecated. "
                "Consider using the new `layout` keyword instead.",
                PendingDeprecationWarning,
            )

        if "layout" in self.configuration and not isinstance(
            self.configuration["layout"], list
        ):
            raise ParserError(
                f"{terminal_colors.RED}{terminal_colors.BOLD}"
                "The configuration input belonging to the "
                "`layout` keyword should be a list."
                f"{terminal_colors.END}"
            )

        if "pageContents" not in self.configuration:
            self.configuration["pageContents"] = []

        self._parse_navigation()

    @property
    def configuration(self) -> dict:
        return self._configuration

    @property
    def shared_settings(self) -> dict:
        return self._shared_settings

    @property
    def assets(self) -> set:
        return self._assets

    @property
    def plugin_metadata(self) -> Dict[str, dict]:
        """Returns a dictionary of plugin metadata, and only for
        plugins included in the configuration file.
        """
        return self._plugin_metadata

    def _recursively_parse_navigation_item(self, items: dict, level: int) -> list:
        navigation_items: list = []
        for item_number, item in enumerate(items):
            if "section" in item:
                if level > 0:
                    raise ParserError(
                        f"{terminal_colors.RED}{terminal_colors.BOLD}"
                        "Sections can only be defined on the first level"
                        f"{terminal_colors.END}"
                    )
                navigation_items.append(
                    {
                        "type": "section",
                        "title": item["section"],
                        "icon": item["icon"] if "icon" in item.keys() else None,
                        "content": self._recursively_parse_navigation_item(
                            item["content"], level + 1
                        )
                        if "content" in item.keys()
                        else [],
                    }
                )
            elif "group" in item:
                navigation_items.append(
                    {
                        "type": "group",
                        "title": item["group"],
                        "icon": item["icon"] if "icon" in item.keys() else None,
                        "content": self._recursively_parse_navigation_item(
                            item["content"], level + 1
                        )
                        if "content" in item.keys()
                        else [],
                    }
                )
            elif "page" in item or "title" in item:
                page_title = item["page"] if "page" in item else item["title"]
                self._page_titles.append(page_title)
                if "id" not in item:
                    item["id"] = self._generate_page_id(page_title)
                elif item["id"] in self._page_ids:
                    raise ParserError(
                        f"{terminal_colors.RED}{terminal_colors.BOLD}"
                        "You have more than one page "
                        "with the same `id`."
                        f"{terminal_colors.END}"
                    )

                self._page_ids.append(item["id"])

                if "content" not in item:
                    item["content"] = []
                elif not isinstance(item["content"], list):
                    raise ParserError(
                        f"{terminal_colors.RED}{terminal_colors.BOLD}"
                        "The content of page number "
                        f"{item_number + 1} should be a list."
                        f"{terminal_colors.END}"
                    )

                self.configuration["pageContents"].append(item)

                navigation_items.append(
                    {
                        "type": "page",
                        "title": page_title,
                        "href": "/" + item["id"],
                        "icon": item["icon"] if "icon" in item.keys() else None,
                    }
                )

                plugins = [e for e in item["content"] if isinstance(e, dict)]
                for plugin in plugins:
                    plugin_name = next(iter(plugin))
                    plugin_variables = next(iter(plugin.values()))
                    kwargs = {} if plugin_variables is None else {**plugin_variables}

                    if plugin_name not in ConfigParser.STANDARD_PLUGINS:
                        raise ParserError(
                            f"{terminal_colors.RED}{terminal_colors.BOLD}"
                            "You have included a plugin with "
                            f"name `{plugin_name}` in your "
                            "configuration file. This is not a "
                            "standard plugin."
                            f"{terminal_colors.END}"
                        )

                    plugin["_call_signature"] = _call_signature(
                        plugin_name,
                        kwargs,
                        self._config_folder,
                    )

                    self._assets.update(
                        getattr(webviz_config.plugins, plugin_name).ASSETS
                    )
                    self._plugin_metadata[
                        plugin_name
                    ] = webviz_config.plugins.PLUGIN_METADATA[plugin_name]
        return navigation_items

    def _parse_navigation(self) -> None:
        # pylint: disable=too-many-branches
        """Returns a list of navigation items"""

        if "layout" in self.configuration:
            self.configuration[
                "navigation_items"
            ] = self._recursively_parse_navigation_item(self.configuration["layout"], 0)
        elif "pages" in self.configuration:
            self.configuration[
                "navigation_items"
            ] = self._recursively_parse_navigation_item(self.configuration["pages"], 0)

        options_found = False
        if "options" in self.configuration:
            if "menu" in self.configuration["options"]:
                options_found = True

                if "bar_position" not in self.configuration["options"]["menu"]:
                    self.configuration["options"]["menu"]["bar_position"] = "left"
                elif self.configuration["options"]["menu"]["bar_position"] not in [
                    "left",
                    "top",
                    "right",
                    "bottom",
                ]:
                    raise ParserError(
                        f"{terminal_colors.RED}{terminal_colors.BOLD}"
                        "Invalid option for options > menu > bar_position: "
                        f"{self.configuration['options']['menu']['bar_position']}. "
                        "Please select one of the following options: left, top, right, bottom."
                        f"{terminal_colors.END}"
                    )

                if "drawer_position" not in self.configuration["options"]["menu"]:
                    self.configuration["options"]["menu"]["drawer_position"] = "left"
                elif self.configuration["options"]["menu"]["drawer_position"] not in [
                    "left",
                    "right",
                ]:
                    raise ParserError(
                        f"{terminal_colors.RED}{terminal_colors.BOLD}"
                        "Invalid option for options > menu > drawer_position: "
                        f"{self.configuration['options']['menu']['drawer_position']}. "
                        "Please select one of the following options: left, right."
                        f"{terminal_colors.END}"
                    )

                if "initially_pinned" not in self.configuration["options"]["menu"]:
                    self.configuration["options"]["menu"]["initially_pinned"] = False
                elif not isinstance(
                    self.configuration["options"]["menu"]["initially_pinned"], bool
                ):
                    raise ParserError(
                        f"{terminal_colors.RED}{terminal_colors.BOLD}"
                        "Invalid option for options > menu > initially_pinned: "
                        f"{self.configuration['options']['menu']['initially_pinned']}. "
                        "Please select a boolean value: True, False"
                        f"{terminal_colors.END}"
                    )

                if "homepage" not in self.configuration["options"]["menu"]:
                    self.configuration["options"]["menu"]["homepage"] = None
                elif not isinstance(
                    self.configuration["options"]["menu"]["homepage"], str
                ):
                    raise ParserError(
                        f"{terminal_colors.RED}{terminal_colors.BOLD}"
                        "Invalid option for options > menu > homepage: "
                        f"{self.configuration['options']['menu']['homepage']}. "
                        "Please select a valid string value"
                        f"{terminal_colors.END}"
                    )
                elif (
                    self.configuration["options"]["menu"]["homepage"]
                    not in self._page_titles
                ):
                    raise ParserError(
                        f"{terminal_colors.RED}{terminal_colors.BOLD}"
                        "Invalid option for options > menu > homepage: "
                        f"{self.configuration['options']['menu']['homepage']}. "
                        f"Please check your config file and use the name of an existing page."
                        f"{terminal_colors.END}"
                    )
                else:
                    self.configuration["options"]["menu"]["homepage"] = self._page_ids[
                        self._page_titles.index(
                            self.configuration["options"]["menu"]["homepage"]
                        )
                    ]

                if "initially_collapsed" not in self.configuration["options"]["menu"]:
                    self.configuration["options"]["menu"]["initially_collapsed"] = False
                elif not isinstance(
                    self.configuration["options"]["menu"]["initially_collapsed"], bool
                ):
                    raise ParserError(
                        f"{terminal_colors.RED}{terminal_colors.BOLD}"
                        "Invalid option for options > menu > initially_collapsed: "
                        f"{self.configuration['options']['menu']['initially_collapsed']}. "
                        "Please select a boolean value: True, False"
                        f"{terminal_colors.END}"
                    )

                if "show_logo" not in self.configuration["options"]["menu"]:
                    self.configuration["options"]["menu"]["show_logo"] = True
                elif not isinstance(
                    self.configuration["options"]["menu"]["show_logo"], bool
                ):
                    raise ParserError(
                        f"{terminal_colors.RED}{terminal_colors.BOLD}"
                        "Invalid option for options > menu > show_logo: "
                        f"{self.configuration['options']['menu']['show_logo']}. "
                        "Please select a boolean value: True, False"
                        f"{terminal_colors.END}"
                    )

        if not options_found:
            if "options" not in self.configuration:
                self.configuration["options"] = {}

            self.configuration["options"]["menu"] = {
                "bar_position": "left",
                "drawer_position": "left",
                "initially_pinned": False,
                "initially_collapsed": False,
                "show_logo": True,
                "homepage": None,
            }



import copy
import pathlib
from typing import Any

from ._build_docs import get_plugin_documentation


JSON_SCHEMA = {
    "$id": "https://github.com/equinor/webviz-config",
    "$schema": "http://json-schema.org/draft-07/schema#",
    "title": "Webviz",
    "type": "object",
    "properties": {
        "title": {"description": "Title of your Webviz application.", "type": "string"},
        "shared_settings": {"type": "object"},
        "options": {
            "type": "object",
            "menu": {
                "description": "Define the menu options.",
                "type": "object",
                "properties": {
                    "show_logo": {
                        "description": "State if a logo shall be shown in the menu.",
                        "type": "boolean",
                    },
                    "bar_position": {
                        "description": "Define where the menu bar shall be positioned:"
                        " left, top, right, bottom.",
                        "type": "string",
                    },
                    "drawer_position": {
                        "description": "Define where the menu drawer shall be positioned:"
                        " left or right.",
                        "type": "string",
                    },
                    "initially_pinned": {
                        "description": "State if the menu shall be pinned when initially showing.",
                        "type": "boolean",
                    },
                    "initially_collapsed": {
                        "description": "State if all groups in menu shall initially be collapsed",
                        "type": "boolean",
                    },
                    "homepage": {
                        "description": """
                        Set a custom page as homepage to which the user returns when clicking on the logo. 
                        Use the page's title.
                        """,
                        "type": "string",
                    },
                },
                "additionalProperties": False,
            },
        },
        "layout": {
            "description": "Define the pages (and potential sections and groups)"
            " in your Webviz application.",
            "type": "array",
            "minLength": 1,
            "items": {
                "oneOf": [
                    {
                        "type": "object",
                        "properties": {
                            "section": {
                                "description": "Title of the section.",
                                "type": "string",
                            },
                            "icon": {
                                "description": "Optionally set an icon for the section.",
                                "type": "string",
                            },
                            "content": {
                                "description": "Define the pages (and potential subgroups)"
                                " of this section.",
                                "type": "array",
                                "minLength": 1,
                                "items": {
                                    "type": "object",
                                    "oneOf": [
                                        {"$ref": "#/properties/layout/items/oneOf/1"},
                                        {"$ref": "#/properties/layout/items/oneOf/2"},
                                    ],
                                },
                            },
                        },
                        "required": ["section", "content"],
                        "additionalProperties": False,
                    },
                    {
                        "type": "object",
                        "properties": {
                            "group": {
                                "description": "Title of the group.",
                                "type": "string",
                            },
                            "icon": {
                                "description": "Optionally define an icon for the group.",
                                "type": "string",
                            },
                            "content": {
                                "description": "Content of the group.",
                                "type": "array",
                                "minLength": 1,
                                "items": {
                                    "oneOf": [
                                        {"$ref": "#/properties/layout/items/oneOf/1"},
                                        {"$ref": "#/properties/layout/items/oneOf/2"},
                                    ]
                                },
                            },
                        },
                        "required": ["group", "content"],
                        "additionalProperties": False,
                    },
                    {
                        "type": "object",
                        "properties": {
                            "page": {
                                "description": "Title of the page.",
                                "type": "string",
                            },
                            "icon": {
                                "description": "Optionally define an icon for the page.",
                                "type": "string",
                            },
                            "content": {
                                "description": "Content of the page.",
                                "type": "array",
                                "minLength": 1,
                                "items": {"oneOf": [{"type": "string"}]},
                            },
                        },
                        "required": ["page", "content"],
                        "additionalProperties": False,
                    },
                ],
            },
        },
    },
    "required": ["title", "layout"],
    "additionalProperties": False,
}

CONTACT_PERSON = {
    "contact_person": {
        "type": "object",
        "properties": {
            "name": {"type": "string"},
            "email": {"type": "string"},
            "phone": {"type": ["integer", "string"]},
        },
        "additionalProperties": False,
    }
}


def create_schema() -> dict:
    def json_type(typehint: Any) -> dict:
        # pylint: disable=too-many-return-statements
        if typehint == list:
            return {"type": "array", "items": {}}
        if typehint in (str, pathlib.Path):
            return {"type": "string"}
        if typehint == bool:
            return {"type": "boolean"}
        if typehint == int:
            return {"type": "integer"}
        if typehint == float:
            return {"type": "number"}
        if typehint == dict:
            return {"type": "object"}
        return {}

    json_schema = copy.deepcopy(JSON_SCHEMA)

    # fmt: off
    content_schemas = json_schema["properties"]["layout"][  # type: ignore
        "items"]["oneOf"][2]["properties"]["content"]["items"]["oneOf"]
    # fmt: on

    for package_doc in get_plugin_documentation().values():
        for plugin_doc in package_doc["plugins"]:
            content_schemas.append(
                {
                    "type": "object",
                    "properties": {
                        plugin_doc["name"]: {
                            "description": plugin_doc["description"],
                            "type": "object",
                            "properties": {
                                **{
                                    arg_name: json_type(arg_info.get("typehint"))
                                    for arg_name, arg_info in plugin_doc[
                                        "arg_info"
                                    ].items()
                                },
                                **CONTACT_PERSON,
                            },
                            "required": [
                                arg_name
                                for arg_name, arg_info in plugin_doc["arg_info"].items()
                                if arg_info["required"]
                            ],
                            "additionalProperties": False,
                        }
                    },
                    "required": [plugin_doc["name"]],
                    "additionalProperties": False,
                    "minProperties": 1,
                    "maxProperties": 1,
                }
            )

    return json_schema



import sys
import time
import shutil
import pathlib
import tempfile
import subprocess  # nosec
import argparse

from yaml import YAMLError

from ._config_parser import ParserError
from ._write_script import write_script
from ._dockerize import create_docker_setup
from .themes import installed_themes
from .utils import terminal_colors

BUILD_FILENAME = "webviz_app.py"
STATIC_FOLDER = pathlib.Path(__file__).resolve().parent / "static"


# Restructure here before merge.
# pylint: disable=too-many-branches
def build_webviz(args: argparse.Namespace) -> None:

    if args.theme not in installed_themes:
        raise ValueError(f"Theme `{args.theme}` is not installed.")

    if args.portable is None:
        build_directory = pathlib.Path(tempfile.mkdtemp())
    else:
        build_directory = args.portable.resolve()
        build_directory.mkdir(parents=True)

    shutil.copytree(STATIC_FOLDER / "assets", build_directory / "resources" / "assets")

    for asset in installed_themes[args.theme].assets:
        shutil.copy(asset, build_directory / "resources" / "assets")

    (build_directory / "theme_settings.json").write_text(
        installed_themes[args.theme].to_json()
    )

    try:
        if args.portable:
            print(
                f"{terminal_colors.BLUE}{terminal_colors.BOLD}"
                "Saving requested data to build folder "
                "such that the webviz instance is portable."
                f"{terminal_colors.END}"
            )

            write_script(
                args, build_directory, "copy_data_template.py.jinja2", "copy_data.py"
            )

            if subprocess.call(  # nosec
                [sys.executable, "copy_data.py"], cwd=build_directory
            ):
                # We simply exit here with a non-zero status in order to not clutter
                # the subprocess traceback with unnecessary information
                sys.exit(1)

            (build_directory / "copy_data.py").unlink()

            print(
                f"{terminal_colors.GREEN}{terminal_colors.BOLD}"
                "Finished data extraction. All output saved."
                f"{terminal_colors.END}"
            )

        non_default_assets, plugin_metadata = write_script(
            args, build_directory, "webviz_template.py.jinja2", BUILD_FILENAME
        )

        for asset in non_default_assets:
            shutil.copy(asset, build_directory / "resources" / "assets")

        if args.portable:
            for filename in ["README.md", ".dockerignore", ".gitignore"]:
                shutil.copy(STATIC_FOLDER / filename, build_directory)
            create_docker_setup(build_directory, plugin_metadata)
        else:
            run_webviz(args, build_directory)

    finally:
        if not args.portable:
            shutil.rmtree(build_directory)


def run_webviz(args: argparse.Namespace, build_directory: pathlib.Path) -> None:

    print(
        f"{terminal_colors.YELLOW}"
        " Starting up your webviz application. Please wait..."
        f"{terminal_colors.END}"
    )

    with subprocess.Popen(  # nosec
        [sys.executable, BUILD_FILENAME], cwd=build_directory
    ) as app_process:

        lastmtime = args.yaml_file.stat().st_mtime

        while app_process.poll() is None:
            try:
                time.sleep(1)

                if lastmtime != args.yaml_file.stat().st_mtime:
                    lastmtime = args.yaml_file.stat().st_mtime
                    write_script(
                        args,
                        build_directory,
                        "webviz_template.py.jinja2",
                        BUILD_FILENAME,
                    )
                    print(
                        f"{terminal_colors.BLUE}{terminal_colors.BOLD}"
                        " Rebuilt webviz dash app from configuration file"
                        f"{terminal_colors.END}"
                    )

            except (ParserError, YAMLError) as excep:
                print(
                    f"{excep} {terminal_colors.RED}{terminal_colors.BOLD}"
                    "Fix the error and save the configuration file in "
                    "order to trigger a new rebuild."
                    f"{terminal_colors.END}"
                )

            except KeyboardInterrupt:
                app_process.kill()
                print(
                    f"\r{terminal_colors.BLUE}{terminal_colors.BOLD}"
                    " Shutting down the webviz application on user request."
                    f"{terminal_colors.END}"
                )

            except Exception as excep:
                app_process.kill()
                print(
                    f"{terminal_colors.RED}{terminal_colors.BOLD}"
                    "Unexpected error. Killing the webviz application process."
                    f"{terminal_colors.END}"
                )
                raise excep



from typing import Dict, List, Optional, Tuple, Type, Union
import io
import abc
import base64
import zipfile
import warnings
import sys
import urllib
import enum

import bleach
from dash.development.base_component import Component
from dash import callback, callback_context, dcc, html, Input, Output
import dash
import jinja2
import webviz_core_components as wcc

from .webviz_plugin_subclasses import SettingsGroupABC, ViewABC, LayoutUniqueId

if sys.version_info >= (3, 8):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict


class ZipFileMember(TypedDict):
    filename: str  # Filename to be given within the archive
    content: str  # String representing file content


class EncodedFile(ZipFileMember):
    # Same keys as in ZipFileMember, with the following changes:
    # - filename is now name of the actual downloaded file.
    # - content is now a base64 encoded ASCII string.
    # - mime_type needs to be added as well.
    mime_type: str


def _create_feedback_text(
    plugin_name: str, dist_name: str, dist_version: str, dependencies: Dict[str, str]
) -> str:
    template_environment = jinja2.Environment(  # nosec
        loader=jinja2.PackageLoader("webviz_config", "templates"),
        undefined=jinja2.StrictUndefined,
        autoescape=False,
    )
    template = template_environment.get_template("feedback.md.jinja2")
    return template.render(
        {
            "plugin_name": plugin_name,
            "dist_name": dist_name,
            "dist_version": dist_version,
            "dependencies": dependencies,
        }
    )


class DuplicatePluginChildId(Exception):
    pass


class WebvizPluginABC(abc.ABC):
    # pylint: disable=too-many-public-methods
    """All webviz plugins need to subclass this abstract base class,
    e.g.

    ```python
    class MyPlugin(WebvizPluginABC):

        def __init__(self):
            ...

        def layout(self):
            ...
    ```
    """

    # This is the default set of buttons to show in the rendered plugin
    # toolbar. If the list is empty, the subclass plugin layout will be
    # used directly, without any visual encapsulation layout from this
    # abstract base class. The plugins subclassing this abstract base class
    # can override this variable setting by defining a class constant with
    # the same name.
    #
    # Some buttons will only appear if in addition necessary data is available.
    # E.g. download of zip archive will only appear if the plugin also
    # has defined the corresponding callback, and contact person will only
    # appear if the user configuration file has this information.

    # List of plugin specific assets which should be copied
    # over to the ./assets folder in the generated webviz app.
    # This is typically custom JavaScript and/or CSS files.
    # All paths in the returned ASSETS list should be absolute.
    ASSETS: list = []

    # Counter for instances of the same plugin.
    # Used to create unique ids of DOM elements
    CLASS_INSTANCE_COUNTER: int = 0

    class StorageType(enum.Enum):
        MEMORY = "memory"
        LOCAL = "local"
        SESSION = "session"

    def __init__(
        self,
        screenshot_filename: str = "webviz-screenshot.png",
        stretch: bool = False,
    ) -> None:
        """If a plugin/subclass defines its own `__init__` function
        (which they usually do), they should remember to call
        ```python
        super().__init__()
        ```
        in its own `__init__` function in order to also run the parent `__init__`.
        """
        self.__class__.CLASS_INSTANCE_COUNTER += 1

        self._plugin_unique_id = LayoutUniqueId(
            plugin_uuid=f"{type(self).__name__}-{self.__class__.CLASS_INSTANCE_COUNTER}"
        )
        self._screenshot_filename = screenshot_filename
        self._add_download_button = False

        self._views: List[Tuple[str, ViewABC]] = []
        self._stores: List[Tuple[str, WebvizPluginABC.StorageType]] = []
        self._shared_settings_groups: List[SettingsGroupABC] = []
        self._registered_ids: List[str] = []

        self._active_view_id = ""
        self._stretch = stretch
        self._all_callbacks_set = False

        self._legacy_plugin_view_id = f"{self._plugin_unique_id.to_string()}-view"

        self._set_wrapper_callbacks()

    def uuid(self, element: Optional[str] = None) -> str:
        """Typically used to get a unique ID for some given element/component in
        a plugins layout. If the element string is unique within the plugin, this
        function returns a string which is guaranteed to be unique also across the
        application (even when multiple instances of the same plugin is added).

        Within the same plugin instance, the returned uuid is the same for the same
        element string. I.e. storing the returned value in the plugin is not necessary.

        Main benefit of using this function instead of creating a UUID directly,
        is that the abstract base class can in the future provide IDs that
        are consistent across application restarts (i.e. when the webviz configuration
        file changes in a non-portable setting).
        """

        if element is None:
            return f"{self._plugin_unique_id.get_plugin_uuid()}"

        return f"{element}-{self._plugin_unique_id.get_plugin_uuid()}"

    def set_stretch(self, stretch: bool) -> None:
        self._stretch = stretch

    @property
    def layout(self) -> Union[str, Type[Component]]:
        """This is the only required function of a Webviz plugin.
        It returns a Dash layout which by webviz-config is added to
        the main Webviz application.
        """
        raise NotImplementedError

    def _check_and_register_id(self, id_or_list_of_ids: Union[str, List[str]]) -> None:
        for i in list(
            id_or_list_of_ids
            if isinstance(id_or_list_of_ids, list)
            else [id_or_list_of_ids]
        ):
            if i in self._registered_ids:
                raise DuplicatePluginChildId(
                    f"Duplicate ID in plugin '{type(self).__name__}' detected: '{i}'"
                )
            self._registered_ids.append(i)

    def add_view(self, view: ViewABC, view_id: str, view_group: str = "") -> None:
        # pylint: disable=protected-access
        view.get_unique_id().set_view_id(view_id)
        view._set_get_plugin_shared_settings_func(self.shared_settings_groups)
        view._set_plugin_register_id_func(self._check_and_register_id)
        view._set_plugin_get_store_unique_id_func(self.get_store_unique_id)
        view._set_unique_id(self._plugin_unique_id)
        self._views.append((view_group, view))

    def add_shared_settings_group(
        self,
        settings_group: SettingsGroupABC,
        settings_group_id: str,
        visible_in_views: Optional[List[str]] = None,
        not_visible_in_views: Optional[List[str]] = None,
    ) -> None:
        settings_group.get_unique_id().set_settings_group_id(settings_group_id)
        # pylint: disable=protected-access
        settings_group._set_visible_in_views(
            visible_in_views if visible_in_views else []
        )
        settings_group._set_not_visible_in_views(
            not_visible_in_views if not_visible_in_views else []
        )
        settings_group._set_plugin_register_id_func(self._check_and_register_id)
        settings_group._set_plugin_get_store_unique_id_func(self.get_store_unique_id)
        settings_group._set_unique_id(self._plugin_unique_id)
        self._shared_settings_groups.append(settings_group)

    def add_store(self, store_id: str, storage_type: StorageType) -> None:
        self._stores.append((store_id, storage_type))

    def get_store_unique_id(self, store_id: str) -> str:
        store = next((el[0] for el in self._stores if el[0] == store_id))
        if store:
            return self.uuid(f"store-{store}")

        raise LookupError(
            f"Invalid store id: '{store_id}. Available store ids: {[el[0] for el in self._stores]}"
        )

    def _set_callbacks(self) -> None:
        pass

    def _set_all_callbacks(self) -> None:
        if not self._all_callbacks_set:
            # pylint: disable=protected-access
            self._set_callbacks()
            for view in self._views:
                view[1]._set_all_callbacks()

            for settings_group in self._shared_settings_groups:
                settings_group.set_callbacks()

            self._all_callbacks_set = True

    @property
    def active_view_id(self) -> str:
        return self._active_view_id

    def set_active_view_id(self, view_id: str) -> None:
        view = self.view(view_id)
        if view:
            self._active_view_id = view.get_unique_id().to_string()

    def views(self, view_group: str = "") -> List[Tuple[str, ViewABC]]:
        if view_group != "":
            return list(filter(lambda x: x[0] == view_group, self._views))
        return self._views

    def view(self, view_id: str) -> ViewABC:
        view = next(
            (
                el[1]
                for el in self.views()
                if el[1].get_unique_id().get_view_id() == view_id
            ),
            None,
        )
        if view:
            return view

        raise LookupError(
            f"""
            Invalid view id: '{view_id}. Available view ids: {[
                el[1].get_unique_id().get_view_id() for el in self._views
            ]}
            """
        )

    def unverified_view_uuid(self, view_id: str) -> str:
        view_uuid = LayoutUniqueId(self._plugin_unique_id.get_plugin_uuid(), view_id)
        return view_uuid.to_string()

    def unverified_settings_group_uuid(self, group_id: str) -> str:
        group_uuid = LayoutUniqueId(self._plugin_unique_id.get_plugin_uuid(), group_id)
        return group_uuid.to_string()

    def shared_settings_groups(self) -> List[SettingsGroupABC]:
        return self._shared_settings_groups

    def shared_settings_group(self, settings_group_id: str) -> SettingsGroupABC:
        group = next(
            (
                el
                for el in self.shared_settings_groups()
                if el.get_unique_id().get_settings_group_id() == settings_group_id
            ),
            None,
        )
        if group:
            return group

        raise LookupError(
            f"""
            Invalid settings group id: '{settings_group_id}'.
            Available settings group ids: {[
                    el.get_unique_id().get_settings_group_id()
                    for el in self._shared_settings_groups
                ]}
            """
        )

    def get_all_settings(self) -> List[html.Div]:
        # pylint: disable=protected-access
        settings = []
        shared_settings = self.shared_settings_groups()
        if shared_settings is not None:
            settings = [
                setting._wrapped_layout("", self._plugin_unique_id.to_string())
                for setting in shared_settings
            ]

        for view in self._views:
            settings.extend(
                [
                    setting._wrapped_layout(
                        view[1].unique_id(), self._plugin_unique_id.to_string()
                    )
                    for setting in view[1].settings_groups()
                ]
            )

        return settings

    @property
    def plugin_data_output(self) -> Output:
        self._add_download_button = True
        return Output(self._legacy_plugin_view_id, "download")

    @property
    def plugin_data_requested(self) -> Input:
        return Input(self._legacy_plugin_view_id, "data_requested")

    @staticmethod
    def _reformat_tour_steps(steps: List[dict]) -> List[dict]:
        return [
            {
                "elementId": str(step["id"]),
                "viewId": ""
                if isinstance(step["id"], str)
                else step["id"].get_view_unique_id()
                if step["id"].get_view_unique_id() is not None
                else "",
                "settingsGroupId": ""
                if isinstance(step["id"], str)
                else step["id"].get_settings_group_unique_id()
                if step["id"].is_settings_group()
                else None,
                "viewElementId": ""
                if isinstance(step["id"], str)
                else step["id"].get_view_element_unique_id()
                if step["id"].is_view_element() and step["id"].is_settings_group()
                else None,
                "content": step["content"],
            }
            for step in steps
        ]

    @staticmethod
    def plugin_compressed_data(
        filename: str, content: List[ZipFileMember]
    ) -> EncodedFile:
        with io.BytesIO() as bytes_io:
            with zipfile.ZipFile(bytes_io, "w") as zipped_data:
                for data in content:
                    zipped_data.writestr(data["filename"], data["content"])
            return {
                "filename": filename,
                "content": base64.b64encode(bytes_io.getvalue()).decode("ascii"),
                "mime_type": "application/zip",
            }

    @staticmethod
    def plugin_data_compress(content: List[ZipFileMember]) -> EncodedFile:
        warnings.warn(
            "Use 'plugin_compressed_data' instead of 'plugin_data_compress'",
            DeprecationWarning,
        )
        return WebvizPluginABC.plugin_compressed_data("webviz-data.zip", content)

    def _make_extended_deprecation_warnings(
        self,
        plugin_deprecation_warnings: Optional[List[str]] = None,
        argument_deprecation_warnings: Optional[List[str]] = None,
    ) -> List[Dict[str, str]]:
        # pylint: disable=import-outside-toplevel
        from .plugins import PLUGIN_METADATA, PLUGIN_PROJECT_METADATA

        extended_deprecation_warnings: List[Dict[str, str]] = []

        plugin_name = self.__class__.__name__
        dist_name = PLUGIN_METADATA[plugin_name]["dist_name"]
        metadata = PLUGIN_PROJECT_METADATA[dist_name]

        if plugin_deprecation_warnings:
            url = (
                f"{metadata['documentation_url']}/#/plugin_deprecations?id={plugin_name.lower()}"
                if metadata["documentation_url"] is not None
                else ""
            )

            for plugin_deprecation_warning in plugin_deprecation_warnings:
                extended_deprecation_warnings.append(
                    {"message": plugin_deprecation_warning, "url": url}
                )
        if argument_deprecation_warnings:
            url = (
                f"{metadata['documentation_url']}/#/argument_deprecations?id={plugin_name.lower()}"
                if metadata["documentation_url"] is not None
                else ""
            )

            for argument_deprecation_warning in argument_deprecation_warnings:
                extended_deprecation_warnings.append(
                    {"message": argument_deprecation_warning, "url": url}
                )
        return extended_deprecation_warnings

    def _make_feedback_url(self) -> str:
        # pylint: disable=import-outside-toplevel
        from .plugins import PLUGIN_METADATA, PLUGIN_PROJECT_METADATA

        plugin_name = self.__class__.__name__
        dist_name = PLUGIN_METADATA[plugin_name]["dist_name"]
        metadata = PLUGIN_PROJECT_METADATA[dist_name]

        feedback_url = ""
        if metadata["tracker_url"] is not None:
            queries = {
                "title": "",
                "body": _create_feedback_text(
                    plugin_name,
                    dist_name,
                    metadata["dist_version"],
                    metadata["dependencies"],
                ),
            }

            feedback_url = (
                f"{metadata['tracker_url']}/new?{urllib.parse.urlencode(queries)}"
                if "github.com" in metadata["tracker_url"]
                else metadata["tracker_url"]
            )

        return feedback_url

    def plugin_layout(
        self,
        contact_person: Optional[dict] = None,
        plugin_deprecation_warnings: Optional[List[str]] = None,
        argument_deprecation_warnings: Optional[List[str]] = None,
    ) -> List[Component]:
        """This function returns plugin layout placed within a WebvizPluginWrapper
        component, which contains a settings drawer with useful buttons like fullscreen
        and screenshot of plugin content. Additional buttons as plugin author contact
        information, download of data, guided tour, issue feedback link, etc. will appear
        based on provided information.

        Button for download of data will only appear if corresponding
        callback functionality is implemented for the respective plugin.
        """

        self._set_all_callbacks()

        if self.active_view_id == "" and len(self.views()) > 0:
            self._active_view_id = self.views()[0][1].get_unique_id().to_string()

        if contact_person:
            # Sanitize the configuration user input
            for key in contact_person:
                contact_person[key] = bleach.clean(str(contact_person[key]))

        return [
            dcc.Store(
                id=self.uuid(f"store-{store[0]}"),
                storage_type=store[1].value,
            )
            for store in self._stores
        ] + [
            wcc.WebvizPluginWrapper(
                id=self._plugin_unique_id.to_string(),
                name=type(self).__name__,
                views=[
                    {
                        "id": view[1].unique_id(),
                        "group": view[0],
                        "name": view[1].name,
                        # pylint: disable=protected-access
                        "showDownload": view[1]._add_download_button,
                    }
                    for view in self.views()
                ]
                if self.views()
                else [
                    {
                        "id": self._legacy_plugin_view_id,
                        "group": "",
                        "name": "",
                        "showDownload": self._add_download_button,
                    }
                ],
                initiallyActiveViewId=self.active_view_id
                if self.views()
                else self._legacy_plugin_view_id,
                contactPerson=contact_person,
                deprecationWarnings=self._make_extended_deprecation_warnings(
                    plugin_deprecation_warnings, argument_deprecation_warnings
                ),
                screenshotFilename=self._screenshot_filename,
                feedbackUrl=self._make_feedback_url(),
                tourSteps=WebvizPluginABC._reformat_tour_steps(self.tour_steps)
                if hasattr(self, "tour_steps")
                else None,
                stretch=self._stretch if self.views() else True,
                children=[
                    wcc.WebvizPluginLoadingIndicator()
                    if self.views()
                    else wcc.WebvizView(
                        id=self._legacy_plugin_view_id,
                        children=[
                            html.Div(
                                children=[self.layout],
                                style={"width": "100%", "margin-left": "16px"},
                            )
                        ],
                    )
                ],
                persistence_type="session",
                persistence=True,
            )
        ]

    def _set_wrapper_callbacks(self) -> None:
        @callback(
            Output(self._plugin_unique_id.to_string(), "children"),
            Input("webviz-content-manager", "activeViewId"),
            Input("webviz-content-manager", "activePluginId"),
        )
        def change_view(view_id: str, plugin_id: str) -> Component:
            ctx = callback_context.triggered
            initial_call = (
                ctx[0]["prop_id"] == "."
                if ctx is not None
                and len(ctx) > 0
                and isinstance(ctx[0], dict)
                and "prop_id" in ctx[0].keys()
                else False
            )

            if plugin_id == self._plugin_unique_id.to_string() or initial_call:
                view = next(
                    (
                        view[1]
                        for view in self.views()
                        if view[1].unique_id() == view_id
                    ),
                    None,
                )
                if view:
                    return view.outer_layout()
            return dash.no_update



from typing import Any, Dict, Optional, Type, TypeVar
import warnings

from .webviz_factory import WebvizFactory
from .webviz_instance_info import WebvizInstanceInfo, WEBVIZ_INSTANCE_INFO

T = TypeVar("T", bound=WebvizFactory)


class WebvizFactoryRegistry:
    """Global registry for factories that allows the actual factory instances
    to be shared between plugins. Also facilitates storage of optional factory
    settings that are read from the YAML config file for later consumption when
    an actual factory gets created and added to the registry.
    The registry is exposed globally through WEBVIZ_FACTORY_REGISTRY below.
    This is also the reason for the two-stage initialization approach. Note that
    the registry instance is useless until the initialize() method has been called.

    Note that this functionality is provisional/experimental, and will not necessarily
    see a deprecation phase and "may deviate from the usual version semantics."
    """

    def __init__(self) -> None:
        self._is_initialized: bool = False
        self._factory_settings_dict: Dict[str, Any] = {}
        self._factories: Dict[Type, WebvizFactory] = {}

    def initialize(
        self,
        factory_settings_dict: Optional[Dict[str, Any]],
    ) -> None:
        """Does the actual initialization of the object instance.
        This function will be called as part of the webviz_app.py / jinja2 template.
        """
        if self._is_initialized:
            raise RuntimeError("Registry already initialized")

        if factory_settings_dict:
            if not isinstance(factory_settings_dict, dict):
                raise TypeError("factory_settings_dict must be of type dict")
            self._factory_settings_dict = factory_settings_dict

        self._is_initialized = True

    def set_factory(self, factory_class: Type[T], factory_obj: T) -> None:
        if not self._is_initialized:
            raise RuntimeError("Illegal access, factory registry is not initialized")

        if not isinstance(factory_obj, factory_class):
            raise TypeError("The type of the factory does not match factory_class")

        self._factories[factory_class] = factory_obj

    def get_factory(self, factory_class: Type[T]) -> Optional[T]:
        if not self._is_initialized:
            raise RuntimeError("Illegal access, factory registry is not initialized")

        if not factory_class in self._factories:
            return None

        factory_obj = self._factories[factory_class]
        if not isinstance(factory_obj, factory_class):
            raise TypeError("The stored factory object has wrong type")

        return factory_obj

    def cleanup_resources_after_plugin_init(self) -> None:
        if not self._is_initialized:
            raise RuntimeError("Illegal access, factory registry is not initialized")

        for factory in self._factories.values():
            factory.cleanup_resources_after_plugin_init()

    @property
    def all_factory_settings(self) -> Dict[str, Any]:
        if not self._is_initialized:
            raise RuntimeError("Illegal access, factory registry is not initialized")

        return self._factory_settings_dict

    @property
    def app_instance_info(self) -> WebvizInstanceInfo:
        warnings.warn(
            "Accessing WebvizInstanceInfo through WebvizFactoryRegistry has been deprecated, "
            "please use global WEBVIZ_INSTANCE_INFO instead",
            DeprecationWarning,
        )

        return WEBVIZ_INSTANCE_INFO


WEBVIZ_FACTORY_REGISTRY = WebvizFactoryRegistry()
